package net.conacom.servletrunner.impl;

/**
 * Copyright 1999,2002 by conacom GmbH, http://www.conacom.de
 * Copyright 1999-2002 Bernd Hohmann, hohmann@harddiskcafe.de
 * All rights reserved.
 * 
 * This software or implementation is free for commercial and non-commercial use.
 * No warranty is given about the functionality.
 */
import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import net.conacom.servletrunner.util.*;
import net.conacom.servletrunner.sessionmanager.*;
/**
 * Provides a way to identify a user across more than one page
 * request or visit to a Web site.
 *
 * <p>The servlet engine uses this interface to create a session
 * between an HTTP client and an HTTP server. The session persists
 * for a specified time period, across more than one connection or
 * page request from the user. A session usually corresponds to one 
 * user, who may visit a site many times. The server can maintain a 
 * session either by using cookies or by rewriting URLs.
 *
 * <p>This interface allows servlets to 
 * <ul>
 * <li>View and manipulate information about a session, such as
 * the session identifier, creation time, or context
 * <li>Bind objects to sessions, allowing you to use an
 * online shopping cart
 * to hold data that persists across multiple user connections
 * </ul>
 *
 * <p> <code>HttpSession</code> defines methods that
 * store these types of data:
 * <ul>
 * <li>Standard session properties, such as a session identifier
 * or session context
 * <li>Data that the application provides, accessed using this
 * interface and stored using a dictionary-like interface
 * </ul>
 *
 * <p>The servlet can obtain data from the session, modify it,
 * and return it to the session, as this example shows:
 *
 * <code>
 * //Get the session object - "request" represents the HTTP servlet request
 * HttpSession session = request.getSession(true);
 *
 * //Get the session data value - an Integer object is read from 
 * //the session, incremented, then written back to the session.
 * //sessiontest.counter identifies values in the session
 * Integer ival = (Integer) session.getValue("sessiontest.counter");
 * if (ival==null) 
 *    ival = new Integer(1);
 * else 
 *   ival = new Integer(ival.intValue() + 1);
 * session.putValue("sessiontest.counter", ival);
 * </code>
 *
 * <p>When an application stores an object in or removes it from a
 * session, the session checks whether the object implements
 * {@link HttpSessionBindingListener}. If it does, 
 * the servlet notifies the object that it has been bound to or unbound 
 * from the session.
 * 
 * <p>An HTTP session represents the server's view of the session.
 * The server considers a session new under any of these conditions:
 * <ul>
 * <li>The client does not yet know about the session
 * <li>The session has not yet begun
 * <li>The client chooses not to join the session, for example,
 * if the server supports only cookies and the client rejects the
 * cookies the server sends
 * </ul>
 * When the session is new, the {@link #isNew} method 
 * returns <code>true</code>.
 *
 * <p>You must write your servlet to handle cases in which
 * the client has not joined a session. The following code
 * example calls <code>isNew</code> to determine whether
 * a session is new. If it is, the server requires the client
 * to start a session by directing the client to a welcome
 * page which the user must enter before gaining access to
 * other pages on the site.
 *
 * <code>
 * //Get the session object - "request" represents the HTTP servlet request
 * HttpSession session = request.getSession(true);
 *
 * //insist that the client starts a session
 * //before access to data is allowed
 * //"response" represents the HTTP servlet response
 * if (session.isNew()) {
 *   response.sendRedirect (welcomeURL);
 * }
 * </code>
 *
 *
 * @author	Various
 * @version	$Version$
 *
 *
 * @see 	HttpSessionBindingListener
 * @see 	HttpSessionContext
 */
/**
 * 25.06.02 ho: API 2.3
 * 24.06.02 ee: API 2.2
 * 22.03.02 ho: creation
 */
public class HttpSession_Impl extends Application implements HttpSession {
	private SessionData sd = null;
	// session-id identifier, can be overwritten by SessionManager
	public static String SESSION_IDENTIFIER = "ssrsessionid";
	//
	private ServletContext srvlContext = null;
/**
 * HttpSession_Impl - Konstruktorkommentar.
 */
public HttpSession_Impl(SessionData sd) {
	super();
	this.sd = sd;
}
/**
 * 
 */
public void dispose() {
	blnDisposed = true;
}
/**
 * API 2.2
 * Returns the object bound with the specified name in this session, or null if no object 
 * is bound under the name.
 *
 * Parameters:
 *    name - a string specifying the name of the object
 * Returns:
 *    the object with the specified name
 * Throws:
 *     IllegalStateException - if this method is called on an invalidated session
 */
public Object getAttribute(String strKey) {
	validateSession();
	return sd.hshValues.get(strKey);
}
/**
 * API 2.2
 * Returns an Enumeration of String objects containing the names of all the 
 * objects bound to this session.
 *
 * Returns:
 *    an Enumeration of String objects specifying the names of all the objects bound to this session
 * Throws:
 *    IllegalStateException - if this method is called on an invalidated session
 */
public Enumeration getAttributeNames() {
	validateSession();
	return sd.hshValues.keys();
}
/**
 * Returns the time when this session was created, measured
 * in milliseconds since midnight January 1, 1970 GMT.
 *
 * @return					a <code>long</code> integer specifying
 * 						when this session was created relative
 *						to 1-1-1970 GMT
 *
 * @exception IllegalStateException		if you attempt to get the session's
 *						creation time after the session has
 *						been invalidated
 */
public long getCreationTime() {
	validateSession();
	return sd.lngCreationTime;
}
/**
 * Returns a string containing the unique identifier assigned 
 * to this session. The identifier is assigned 
 * by the servlet engine and is implementation dependent.
 * 
 * @return				a string specifying the identifier
 *					assigned to this session
 *
 * @exeption IllegalStateException	if the session is invalid
 */
public String getId() {
	validateSession();
	return sd.strID;
}
/**
 * Returns the last time the client sent a request associated with
 * this session, as the number of milliseconds since midnight
 * January 1, 1970 GMT. 
 *
 * <p>Actions that your application takes, such as getting or setting
 * a value associated with the session, do not affect the access
 * time.
 *
 * <p>The last accessed time can help you manage sessions. For example,
 * the sessions can be sorted according to age to optimize some task.
 *
 * @return					a <code>long</code> integer 
 *						representing the last time 
 *						the client sent a request associated 

 *						with this session, expressed in 
 *						milliseconds since 1-1-1970 GMT
 *
 * @exception IllegalStateException		if the session is invalid
 */
public long getLastAccessedTime() {
	validateSession();
	return sd.lngLastAccessedTime;
}
/**
 * Returns the maximum time interval, in seconds, that 
 * the servlet engine will keep this session open between 
 * client requests. You can set the maximum time interval
 * with the <code>setMaxInactiveInterval</code> method.
 *
 *  
 *
 * @return		an integer specifying the number of
 *			seconds this session remains open
 *			between client requests
 *
 * @see		#setMaxInactiveInterval
 */
public int getMaxInactiveInterval() {
	return (int) (sd.lngMaxInactiveInterval / 1000l);
}
/**
 * API 2.3
 * Returns the ServletContext to which this session belongs.
 *
 * Returns:
 *   The ServletContext object for the web application
 */
public ServletContext getServletContext() {
	return srvlContext;
}
/**
 * @deprecated 	As of Version 2.1, this method is
 *			deprecated and has no replacement.
 *			It will be removed in a future
 *			version of the Java Servlet API.
 */
public HttpSessionContext getSessionContext() {
	return null;
}
/**
 * Returns the object bound with the specified name
 * in this session or null if no object of
 * that name exists.
 *
 * @param name				a string specifying the name of the object
 *
 * @return					the object with the specified name
 *
 * @exception IllegalStateException		if the session is invalid
 *
 * @deprecated: use getAttribute(..)
 */
public Object getValue(String strKey) {
	validateSession();
	return sd.hshValues.get(strKey);
}
/**
 * Returns an array containing the names of all the objects
 * bound to this session. This method is useful, for example,
 * when you want to delete all the objects bound to this
 * session.
 *
 * @return					an array of strings specifying the
 *						names of all the objects bound to
 *						this session
 *
 * @exception IllegalStateException		if the session is invalid
 *
 * @deprecated: use getAttributes
 */
public String[] getValueNames() {
	validateSession();
	Vector v = new Vector();
	for (Enumeration e = sd.hshValues.keys(); e.hasMoreElements();) {
		v.addElement(e.nextElement());
	}
	String s[] = new String[v.size()];
	v.copyInto(s);
	v.removeAllElements();
	v = null;
	return s;
}
/**
 *
 * Invalidates this session and unbinds any objects bound
 * to it.
 *
 * @exception IllegalStateException		if the session is already invalid
 *
 */
public void invalidate() {
	validateSession();
	sd.lngLastAccessedTime = -1;
}
/**
 * Returns <code>true</code> if the Web server has created a session
 * but the client has not yet joined. 
 * For example, if the server used only cookie-based sessions, and
 * the client had disabled the use of cookies, then a session would
 * be new.
 *
 * @return 					<code>true</code> if the 
 *						server has created a session, 
 *						but the client has not yet joined
 *
 * @exception IllegalStateException		if the session is invalid
 */
public boolean isNew() {
	validateSession();
	return sd.blnIsNewSession;
}
/**
 * Binds an object to this session, using the name specified.
 * If an object of the same name is already bound to the session,
 * the object is replaced.
 *
 *
 * @param name				the name to which the object is bound;
 *						cannot be null
 *
 * @param value				the object to be bound; cannot be null
 *
 * @exception IllegalStateException		if the session is invalid
 */
public void putValue(String strKey, Object objData) {
	validateSession();
	sd.hshValues.put(strKey, objData);
	if (objData != null && objData instanceof HttpSessionBindingListener)
		 ((HttpSessionBindingListener) objData).valueBound(new HttpSessionBindingEvent(this, strKey));
}
/**
 * API 2.2
 * Removes the object bound with the specified name from this session. If 
 * the session does not have an object bound with the specified name, this method does nothing.
 *
 * After this method executes, and if the object implements HttpSessionBindingListener, the container 
 * calls HttpSessionBindingListener.valueUnbound. The container then notifies any HttpSessionAttributeListeners 
 * in the web application.
 * 
 * Parameters:
 *    name - the name of the object to remove from this session 
 */
public void removeAttribute(String strKey) {
	validateSession();
	Object o = sd.hshValues.remove(strKey);
	if (o instanceof HttpSessionBindingListener) {
		((HttpSessionBindingListener) o).valueUnbound(new HttpSessionBindingEvent(this, strKey));
	}
}
/**
 * Removes the object bound with the specified name from
 * this session. If the session does not have an object
 * bound with the specified name, this method does nothing.
 *
 * <p>After this method executes, and if the object
 * implements <code>HttpSessionBindingListener</code>,
 * the object calls 
 * <code>HttpSessionBindingListener.valueUnbound</code>.
 * 
 * 
 *
 * @param name				the name of the object to
 *						remove from this session
 *
 * @exception IllegalStateException		if the session is invalid
 */
public void removeValue(String strKey) {
	validateSession();
	Object o = sd.hshValues.remove(strKey);
	if (o instanceof HttpSessionBindingListener) {
		((HttpSessionBindingListener) o).valueUnbound(new HttpSessionBindingEvent(this, strKey));
	}
}
/**
 * API 2.2
 * Binds an object to this session, using the name specified. If an object of the 
 * same name is already bound to the session, the object is replaced.
 *
 * After this method executes, and if the new object implements HttpSessionBindingListener, 
 * the container calls HttpSessionBindingListener.valueBound. The container then notifies 
 * any HttpSessionAttributeListeners in the web application.
 *
 * If an object was already bound to this session of this name that implements 
 * HttpSessionBindingListener, its HttpSessionBindingListener.valueUnbound method is called.
 *
 * If the value passed in is null, this has the same effect as calling removeAttribute().
 */
public void setAttribute(String strKey, Object objData) {
	validateSession();
	if (objData == null && sd.hshValues.get(strKey) != null) {
		removeAttribute(strKey);
		return;
	}
	if (objData != null) {
		sd.hshValues.put(strKey, objData);
		if (objData instanceof HttpSessionBindingListener)
			 ((HttpSessionBindingListener) objData).valueBound(new HttpSessionBindingEvent(this, strKey));
	}
}
/**
 * Specifies the maximum length of time, in seconds, that the
 * servlet engine keeps this session if no user requests
 * have been made of the session.
 *
 * @param interval		An integer specifying the number
 * 				of seconds 
 */
public void setMaxInactiveInterval(int intInterval) {
	sd.lngMaxInactiveInterval = intInterval * 1000l;
}
/**
 * checks if the session is valid. this is done by checking the lastAccessedTime
 * sets also the "new session flag"
 */
private void validateSession() {
	if (!sd.isValidSession())
		throw new IllegalStateException("Session " + sd.strID + " is not a valid session");
}
/**
 * returns the changed sessiondata (for SessionStorage class)
 */
public SessionData _getSessionData() {
	return sd;
}
/**
 * HELPER: set the current servlet context for getServletContext()
 */
public void _setServletContext(ServletContext sc) {
	srvlContext = sc;
}
}