package net.conacom.servletrunner.impl;

/**
 * Copyright 1999,2002 by conacom GmbH, http://www.conacom.de
 * Copyright 1999-2002 Bernd Hohmann, hohmann@harddiskcafe.de
 * All rights reserved.
 * 
 * This software or implementation is free for commercial and non-commercial use.
 * No warranty is given about the functionality.
 */
import java.io.*;
import java.util.*;
import javax.servlet.*;
import net.conacom.servletrunner.util.*;
import net.conacom.servletrunner.main.*;
/**
 * Sends MIME-encoded data from the servlet to the client. The servlet
 * engine creates a <code>ServletResponse</code> object and passes it 
 * as an argument to the servlet's <code>service</code> method.
 *
 * <p>To send binary data in a MIME body response, use
 * the {@link ServletOutputStream} returned by {@link #getOutputStream}.
 * Likewise, to send text data, use the <code>PrintWriter</code> object 
 * returned by {@link #getWriter}. If you need to mix binary and text data,
 * for example, if you are creating a multipart response, use a
 * <code>ServletOutputStream</code> to write the multipart headers, and
 * then use the headers to build your own text bodies.
 *
 * <p>If you do not specify a character set for the MIME body response
 * with {@link #setContentType}, the servlet engine will select one
 * and modify the content accordingly. Call <code>setContentType</code>
 * before you call <code>getWriter</code> or <code>getOutputStream</code>.
 * 
 * <p>See the Internet RFCs such as RFC 2045 at 
 * <a href="http://info.internet.isi.edu/in-notes/rfc/files/rfc2045.txt">RFC 2045</a> 
 * for more information on MIME. Protocols such as SMTP
 * and HTTP define profiles of MIME, and those standards
 * are still evolving.
 *
 * @author 	Various
 * @version 	$Version$
 *
 * @see		ServletOutputStream
 */
/**
 * 25.06.02 ho: API 2.3
 * 24.06.02 ee: API 2.2
 * 21.03.02 ho: creation
 */
public class ServletResponse_Impl extends Application implements ServletResponse {
	// the output
	protected ByteString bsResponse = null;
	// table of headers
	protected DataVector dvHeaders = null;
	// a printwriter
	private PrintWriter pwOut = null;
	// a stream
	private ServletOutputStream sosOut = null;
	// the container config
	protected ServletContainerConfig scc = null;
	// committed (means: runner retrieved the data)
	private boolean blnCommitted = false;
	// moved from HttpServletResponse because of the idiots of SUN (needed for reset())
	// status set by sendError(...)
	protected String strStatus = null;
	// status set by sendError(...)
	protected int intStatus = 0;
/**
 * intBufferSize = the default size for the servlet answer. because extending the buffersize on runtime is
 * very expensive, we try to create a matching buffer. 
 * see ServletContainerStatistics
 */
public ServletResponse_Impl(ServletContainerConfig scc) {
	super();
	this.scc = scc;
	this.bsResponse = new ByteString(scc.intOutputBufferSize);
	this.dvHeaders = new DataVector();
}
/**
 * 
 */
public void dispose() {
	blnDisposed = true;
	if (bsResponse != null) {
		bsResponse.dispose();
		bsResponse = null;
	}
}
/**
 * Clears the content of the underlying buffer in the response without clearing headers or 
 * status code. If the response has been committed, this method throws an IllegalStateException. 
 */
/**
 * ho: this does nothing because we're working buffered all the time
 */
public void flushBuffer() {
	if (blnCommitted)
		throw new IllegalStateException("Buffer already commited");
}
/**
 * Returns the actual buffer size used for the response. If no buffering is used, this 
 * method returns 0.
 *
 * Returns:
 *  the actual buffer size used 
 */
public int getBufferSize() {
	return scc.intOutputBufferSize;
}
/**
 * Returns the name of the character set encoding used for
 * the MIME body sent by this response.
 *
 * <p>The character encoding is either the one specified in
 * the content itself or another one the client understands.
 * If no character encoding has been assigned, it is implicitly
 * set to <i>text/plain</i>.
 *
 * <p>See RFC 2047 (http://ds.internic.net/rfc/rfc2045.txt)
 * for more information about character encoding and MIME.
 *
 * @return	a <code>String</code> specifying the
 *			name of the character set encoding, for
 *			example, <i>text/plain</i>
 *
 * ho: this doesn't do anything useful, needs the HttpServletResponse_Impl
 */
public String getCharacterEncoding() {
	if (dvHeaders.containsKey("Content-Type"))
		return (String) dvHeaders.dataForKey("Content-Type");
	else
		return "text/plain";
}
/**
 * Returns the locale assigned to the response. 
 */
/**
 * ho: doesn't do anything useful, needs further examination
 */
public Locale getLocale() {
	return Locale.getDefault();
}
/**
 * Returns a {@link ServletOutputStream} suitable for writing binary 
 * data in the response. The servlet engine does not encode the
 * binary data.
 *
 * @return				a {@link ServletOutputStream} for writing binary data	
 *
 * @exception IllegalStateException if you have already called the <code>getWriter</code> method
 * 					for the same response
 *
 * @exception IOException 		if an input or output exception occurred
 *
 * @see 				#getWriter
 */
public ServletOutputStream getOutputStream() throws IOException {
	if (isCommitted())
		throw new IllegalStateException("Buffer already committed");
	if (pwOut != null)
		throw new IllegalStateException("Stream already requested as PrintWriter");
	if (sosOut == null)
		sosOut = new ServletOutputStream_Impl(bsResponse);
	return sosOut;
}
/**
 * Returns a <code>PrintWriter</code> object that you
 * can use to send character text to the client. 
 * The character encoding used is the one specified 
 * in the <code>charset=</code> property of the
 * {@link #setContentType} method, which you must call
 * <i>before</i> you call this method. 
 *
 * <p>If necessary, the MIME type of the response is 
 * modified to reflect the character encoding used.
 *
 * <p> You cannot use this method if you have already
 * called {@link #getOutputStream} for this 
 * <code>ServletResponse</code> object.
 *
 * 
 * @return 					a <code>PrintWriter</code> object that 
 *						can return text to the client 
 *
 * @exception UnsupportedEncodingException  if the character encoding specified in
 *						<code>setContentType</code> cannot be
 *						used
 *
 * @exception IllegalStateException    	if the <code>getOutputStream</code>
 * 						method has already been called for this 
 *						response object; in that case, you can't
 *						use this method
 *
 * @exception IOException   		if an input or output exception occurred
 *
 * @see 					#getOutputStream
 * @see 					#setContentType
 */
public PrintWriter getWriter() throws IOException {
	if (isCommitted())
		throw new IllegalStateException("Buffer already committed");
	if (sosOut != null)
		throw new IllegalStateException("Stream already requested as Outputstream");
	if (pwOut == null) {
		OutputStream out = new ServletOutputStream_Impl(bsResponse);
		Writer wr = new OutputStreamWriter(out); // ho: need to set the charset here
		pwOut = new PrintWriter(wr, false);
	}
	return pwOut;
}
/**
 * Returns a boolean indicating if the response has been committed. A commited 
 * response has already had its status code and headers written. 
 */
public boolean isCommitted() {
	return blnCommitted;
}
/**
 * Clears any data that exists in the buffer as well as the status code and headers. 
 * If the response has been committed, this method throws an IllegalStateException.
 */
public void reset() {
	resetBuffer();
	strStatus = null;
	intStatus = 0;
}
/**
 * Clears the content of the underlying buffer in the response without clearing headers or 
 * status code. If the response has been committed, this method throws an IllegalStateException.
 */
public void resetBuffer() {
	if (isCommitted())
		throw new IllegalStateException("Buffer already committed");
	Helper.flush(pwOut);
	Helper.flush(sosOut);
	bsResponse.setLength(0);
}
/**
 * Sets the preferred buffer size for the body of the response. The servlet container 
 * will use a buffer at least as large as the size requested. The actual buffer size 
 * used can be found using getBufferSize.
 *
 * A larger buffer allows more content to be written before anything is actually sent, 
 * thus providing the servlet with more time to set appropriate status codes and headers. 
 * A smaller buffer decreases server memory load and allows the client to start receiving 
 * data more quickly.
 *
 * This method must be called before any response body content is written; if content has 
 * been written, this method throws an IllegalStateException.
 *
 * Parameters:
 *   size - the preferred buffer sizeThrows:
 * IllegalStateException - if this method is called after content has been written 
 */
/**
 * ho: the outputbuffersize will never decrease !
 */
public void setBufferSize(int intBufferSize) {
	if (sosOut != null)
		throw new IllegalStateException("Outputbuffer already requested as stream");
	if (pwOut != null)
		throw new IllegalStateException("Outputbuffer already requested as writer");
	bsResponse.setLength(intBufferSize);
	scc.intOutputBufferSize = intBufferSize;
}
/**
 * Sets the length of the content the server returns
 * to the client. In HTTP servlets, this method sets the
 * HTTP Content-Length header.
 *
 *
 * @param len 	an integer specifying the length of the 
 * 			content being returned to the client; sets
 *			the Content-Length header
 *
 * ho: this doesn't do anything useful, needs the HttpServletResponse_Impl
 */
public void setContentLength(int intLength) {
	dvHeaders.insertOrReplaceDataForKey("Content-Length", Integer.toString(intLength));
}
/**
 * Sets the content type of the response the server sends to
 * the client. The content type may include the type of character
 * encoding used, for example, <code>text/html; charset=ISO-8859-4</code>.
 *
 * <p>You can only use this method once, and you should call it
 * before you obtain a <code>PrintWriter</code> or 
 * {@link ServletOutputStream} object to return a response.
 *
 *
 * @param type 	a <code>String</code> specifying the MIME 
 *			type of the content
 *
 * @see 		#getOutputStream
 * @see 		#getWriter
 *
 */
public void setContentType(String strType) {
	if (!dvHeaders.containsKey("Content-Type"))
		dvHeaders.addKeyData("Content-Type", strType);
}
/**
 * Sets the locale of the response, setting the headers (including the Content-Type's charset) 
 * as appropriate. This method should be called before a call to getWriter(). By default, the 
 * response locale is the default locale for the server.
 *
 * Parameters:
 *   loc - the locale of the response
 */
/**
 * ho: doesn't do anything useful, needs further examination
 */
public void setLocale(Locale loc) {
}
/**
 * returns the data for sending to the webserver
 */
public ByteString _getResponseData() {
	Helper.flush(sosOut);
	Helper.close(sosOut);
	Helper.flush(pwOut);
	Helper.close(pwOut);
	blnCommitted = true;
	return bsResponse;
}
/**
 * INTERNAL: get the statusocde
 */
public int _getStatusCode() {
	return intStatus;
}
/**
 * INTERNAL: get the statusmessage
 */
public String _getStatusMessage() {
	return strStatus;
}
/**
 * reinitializes this object
 */
public void _initObject() {
	bsResponse.clear();
	dvHeaders.clear();
	pwOut = null;
	sosOut = null;
	strStatus = null;
	intStatus = 0;
	blnCommitted = false;
}
}