package net.conacom.servletrunner.main;

/**
 * Copyright 1999,2002 by conacom GmbH, http://www.conacom.de
 * Copyright 1999-2002 Bernd Hohmann, hohmann@harddiskcafe.de
 * Copyright 1999 iMatix, http://www.imatix.com
 * All rights reserved.
 * 
 * This software or implementation is free for commercial and non-commercial use.
 * No warranty is given about the functionality.
 */
import java.io.*;
import java.net.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import net.conacom.servletrunner.impl.*;
import net.conacom.servletrunner.util.*;
/**
 * a connection to the webserver using the LRWP 2.0 protocol
 *
 * 25.06.02 ho: headers are string arrays now
 * 10.04.02 ho: some speedups
 * 25.03.02 ho: returning the requestbody as ServletInputStream instead of ByteString
 * 25.03.02 ho: forgot to clear hashtable for parameters between calls
 * 19.03.02 ho: finished
 * 15.03.02 ho: creation
 */
public class LRWP_Connection extends Application {
	// size of LRWP-header
	private final static int HEADERSIZE = 9;
	// buffer for retrieving the length information 
	private byte bytHeaderBuffer[] = new byte[HEADERSIZE];
	// delimiter for LRWP
	private final static byte DELIMITER = (byte) 255;
	// supported LRWP versions
	private final static String SUPPORTED_VERSIONS[] = new String[] {"2.0"};
	// configuration for this connection
	private LRWP_ConnectionConfig cfg = null;
	// our socket
	private Socket sok = null;
	// inputstream of the socket
	private InputStream inp = null;
	// outputstream of the socket
	private OutputStream out = null;
	// universal buffer
	private ByteString bsBuffer = null;
	// hashtable with the header (all parameters)
	private Hashtable hshHeader = null;
	// hashtable with the parameters (from the query-string)
	private Hashtable hshParameters = null;
	// the requestbody (needed for the InputStream feature of the ServletRequest)
	private ByteString bsRequestBody = null;
/**
 *
 */
public LRWP_Connection(LRWP_ConnectionConfig cfg) {
	super();
	this.cfg = cfg;
	init();
}
/**
 * wait for data of the webserver and retrieves the response
 */
public void dataRetrieve() {
	if (sok == null) {
		throw new RunnerException("Not connected to a webserver !");
	}
	// gets the CGI environment
	if (!retrieveHeader())
		return;
	// gets the request-body
	if (!retrieveBody())
		return;
}
/**
 * sends a responsestring to the webserver
 */
public void dataSend(ByteString bs) {
	if (sok == null) {
		throw new RunnerException("Not connected to a webserver !");
	}
	// build a right aligned, 0 padded string with the length of the response
	bsBuffer.clear();
	String s = String.valueOf(bs.getLength());
	bsBuffer.append(s);
	for (int i = s.length(); i < HEADERSIZE; i++) {
		bsBuffer.append('\000');
	}
	// send the data to the server
	try {
		out.write(bsBuffer.getByteArray(), 0, HEADERSIZE);
		out.write(bs.getByteArray(), 0, bs.getLength());
		out.flush();
	} catch (IOException e) {
		dispose();
		throw new RunnerException(e);
	}
}
/**
 * housekeeping
 */
public void dispose() {
	//
	blnDisposed = true;
	//
	if (bsBuffer != null) {
		bsBuffer.dispose();
		bsBuffer = null;
	}
	if (hshHeader != null) {
		hshHeader.clear();
		hshHeader = null;
	}
	if (hshParameters != null) {
		hshParameters.clear();
		hshParameters = null;
	}
	if (bsRequestBody != null) {
		bsRequestBody.dispose();
		bsRequestBody = null;
	}
	// clear sockets
	Helper.close(inp);
	inp = null;
	Helper.flush(out);
	Helper.close(out);
	out = null;
	Helper.close(sok);
	sok = null;
}
/**
 * returns the header (cgi-environment)
 */
public Hashtable getHeader() {
	return hshHeader;
}
/**
 * returns the parameters (parsed QUERY_STRING + parsed requestbody)
 */
public Hashtable getParameters() {
	return hshParameters;
}
/**
 * returns the requestbody as a stream (for HttpReqest#getInputStream()
 */
public ServletInputStream getRequestBody() {
	return new ServletInputStream_Impl(bsRequestBody);
}
/**
 * convience method: every error is rethrown as RunnerException. 
 * we catch it here and do so some housekeeping.
 */
private void init() {
	// init
	bsBuffer = new ByteString(1024);
	bsRequestBody = new ByteString(2048);
	hshHeader = new Hashtable();
	hshParameters = new Hashtable();
	//
	try {
		login();
	} catch (RunnerException e) {
		dispose();
		throw e;
	}
}
/**
 * still connected to the webserver ?
 */
public boolean isConnected() {
	return !blnDisposed && sok != null;
}
/**
 * login to webserver. called from init()
 */
private void login() {
	// check, if we support this LRWP version
	boolean blnVersionSupported = false;
	for (int i = 0; i < SUPPORTED_VERSIONS.length; i++) {
		if (SUPPORTED_VERSIONS[i].equals(cfg.strLRWP_Version)) {
			blnVersionSupported = true;
			break;
		}
	}
	if (!blnVersionSupported) {
		throw new RunnerException("Unsupported LRWP version: " + cfg.strLRWP_Version);
	}

	// open socket, get streams
	int intCnt = 0;
	while (true) {
		try {
			sok = new Socket(cfg.strLRWP_Host, cfg.intLRWP_Port);
			sok.setSoTimeout(0); // never timeout
			sok.setTcpNoDelay(true);
			//
			inp = sok.getInputStream();
			out = sok.getOutputStream();
			// ok, bail out
			break;
		} catch (UnknownHostException e1) {
			throw new RunnerException(e1);
		} catch (IOException e2) {
			// retry limit reached ? bail out with an exception
			if (++intCnt > cfg.intRetries)
				throw new RunnerException(e2);
			cfg.logContext.write("! Retrying LRWP startup " + intCnt + "/" + cfg.intRetries);
			try {
				Thread.currentThread().sleep(cfg.lngRetryWait);
			} catch (InterruptedException e3) {
			}
		}
	}

	// login
	if (cfg.strLRWP_Version.equals("2.0")) {
		login_v200();
	}

	// either we're bailed out by an RunnerException (and disposed socket) or we're ready to go.
}
/**
 * Login to a LRWP Server Version 2.0
 * missing: proper response to a CHALLENGE request
 * not tested: REJECTED
 *
 * called from login()
 */
private void login_v200() {
	/**
	 * send connect request
	 */
	bsBuffer.clear();
	bsBuffer.append(DELIMITER);
	bsBuffer.append(cfg.strLRWP_Version);
	bsBuffer.append(DELIMITER);
	bsBuffer.append(cfg.strServletIdentifier);
	bsBuffer.append(DELIMITER);
	bsBuffer.append(cfg.strVirtualHost);
	bsBuffer.append(DELIMITER);
	try {
		out.write(bsBuffer.getByteArray(), 0, bsBuffer.getLength());
	} catch (IOException e) {
		throw new RunnerException(e);
	}

	/**
	 * read the response into our byte-string 
	 */
	bsBuffer.clear();
	ByteString bs = new ByteString(512); // local buffer only for login
	int intFirstDelimiterPos = -1; // position of the first delimiter
	while (true) {
		try {
			int ii = inp.read(bs.getByteArray());
			if (ii == -1) {
				throw new RunnerException("Connection closed by Webserver");
			}
			bs.setLengthIndex(ii);
		} catch (IOException e) {
			throw new RunnerException(e);
		}
		bsBuffer.append(bs);
		intFirstDelimiterPos = bsBuffer.getIndexOf(DELIMITER);
		// OK+delimiter ? were done.
		if (bsBuffer.startsWith("OK" + (char) DELIMITER)) {
			break;
		}
		// simple OK or ERROR -> got a response from a LRWP 1.0 server !
		if (bsBuffer.startsWith("OK") || bsBuffer.startsWith("ERROR")) {
			throw new RunnerException("Got a response from a LRWP 1.0 server, expecting Version 2.0 !");
		}
		// CHALLENGE ? (NOT IMPLEMENTED YET)
		if (bsBuffer.startsWith("CHALLENGE")) {
			throw new RunnerException("CHALLENGE command not implemented yet");
		}
		// REJECTED ? read full errormessage (NOT TESTED)
		if (bsBuffer.startsWith("REJECTED" + (char) DELIMITER)) {
			// did we read the whole message ?
			if (bsBuffer.getIndexOf(DELIMITER, intFirstDelimiterPos) != -1) {
				String s = bsBuffer.toString();
				throw new RunnerException("REJECTED. Reason:" + s.substring(intFirstDelimiterPos));
			}
		}
	}
	/**
	 * were done.
	 */
}
/**
 * testclass
 */
public static void main(java.lang.String[] args) throws Throwable {
	LRWP_ConnectionConfig cfg = new LRWP_ConnectionConfig();
	cfg.logContext = new LogContext("", "tst", ".xxx");
	cfg.strLRWP_Host = "127.0.0.1";
	cfg.intLRWP_Port = 81;
	cfg.strLRWP_Version = "2.0";
	cfg.strAccessKey = "";
	cfg.strServletIdentifier = "test_appl";
	cfg.strVirtualHost = "bernd.seckbach.conacom.de";
	LRWP_Connection con = new LRWP_Connection(cfg);
	con.dataRetrieve();
	ByteString bs = new ByteString();
	bs.append("Content-type: text/plain\n\n");
	Hashtable hsh = con.getHeader();
	for (Enumeration e = hsh.keys(); e.hasMoreElements();) {
		String strKey = (String) e.nextElement();
		bs.append(strKey + "=" + (String) hsh.get(strKey) + "\n");
	}
	bs.append("[end of data]");
	con.dataSend(bs);
	con.dispose();
}
/**
 * reads the inputstream until the given number of bytes was read
 * returns true if we had read all bytes
 */
private boolean readFully(byte b[], int intLength) {
	try {
		int i = 0;
		while (i < intLength) {
			int cnt = inp.read(b, i, intLength - i);
			if (cnt < 0)
				return false;
			i += cnt;
		}
		return true;
	} catch (IOException e) {
		return false;
	}
}
/**
 * reads a LRWP-string from the webserver
 * returns false if the stream was closed or has error
 */
private boolean readLRWPBuffer(ByteString bs) {
	boolean b = false;
	// retrieve header (9 bytes numeric string indicating the length of the following data)
	b = readFully(bytHeaderBuffer, HEADERSIZE);
	if (!b)
		return false;
	// we're parsing the buffer by hand because its faster and we don't need any char-translation
	int intDataLength = 0;
	for (int i = 0; i < HEADERSIZE; i++) {
		intDataLength *= 10;
		intDataLength += (bytHeaderBuffer[i] - 48);
	}
	// read the buffer
	bs.setLength(intDataLength);
	b = readFully(bs.getByteArray(), intDataLength);
	if (!b)
		return false;
	return true;
}
/**
 * retrieve post-data (if any) and parse it
 */
private boolean retrieveBody() {
	bsRequestBody.clear();
	try {
		if (!readLRWPBuffer(bsRequestBody)) {
			dispose();
			return false;
		}
	} catch (RunnerException e) {
		dispose();
		throw e;
	}

	// check for multipart formdata. in this case we simply return and let
	// the servlet parse it from the inputstream
	String strType[] = (String[]) hshHeader.get("CONTENT_TYPE");
	if (strType != null && strType[0].toLowerCase().startsWith("multipart/form-data")) {
		return true;
	}

	// we have to incoperate the requestbody into the parameter-list
	Hashtable hshTmp = HttpUtils.parseQueryString(bsRequestBody.toString());
	for (Enumeration e = hshTmp.keys(); e.hasMoreElements();) {
		String strKey = (String) e.nextElement();
		hshParameters.put(strKey, hshTmp.get(strKey));
	}
	hshTmp.clear();
	hshTmp = null;
	//
	return true;
}
/**
 * retrieve & parse the environment.
 */
private boolean retrieveHeader() {
	bsBuffer.clear();
	try {
		if (!readLRWPBuffer(bsBuffer)) {
			dispose();
			return false;
		}
	} catch (RunnerException e) {
		dispose();
		throw e;
	}
	// parse the CGI environment
	hshHeader.clear();
	StringTokenizer st = new StringTokenizer(bsBuffer.toString(), "\000");
	while (st.hasMoreTokens()) {
		String strBuffer = st.nextToken();
		int pp = strBuffer.indexOf("=");
		if (pp == -1)
			continue;
		String strLeft = strBuffer.substring(0, pp);
		String strRight = strBuffer.substring(pp + 1);
		if (!hshHeader.containsKey(strLeft)) {
			hshHeader.put(strLeft, new String[] {strRight});
		} else {
			String strCurr[] = (String[]) hshHeader.get(strLeft);
			String strNew[] = new String[strCurr.length + 1];
			for (int i = 0; i < strCurr.length; i++)
				strNew[i] = strCurr[i];
			strNew[strCurr.length + 1] = strRight;
			hshHeader.put(strLeft, strNew);
		}
	}
	// parse the QUERY_STRING from the environment into the parameterlist
	hshParameters.clear();
	String strQuery[] = (String[]) hshHeader.get("QUERY_STRING");
	if (strQuery != null) {
		// don't parse with the SUN parser when
		// no '=' and '&' was found (it will die there...). 
		// but we will try it when we find a '&' (its up to the user to
		// give a correct query-string)
		if (strQuery[0].indexOf('=') == -1 && strQuery[0].indexOf('&') == -1) {
			hshParameters.put(strQuery[0], new String[] {""});
		} else {
			hshParameters = HttpUtils.parseQueryString(strQuery[0]);
		}
	}
	//
	return true;
}
}