package net.conacom.servletrunner.main;

/**
 * Copyright 1999,2002 by conacom GmbH, http://www.conacom.de
 * Copyright 1999-2002 Bernd Hohmann, hohmann@harddiskcafe.de
 * All rights reserved.
 * 
 * This software or implementation is free for commercial and non-commercial use.
 * No warranty is given about the functionality.
 */
import java.io.*;
import java.net.*;
import java.util.*;
import net.conacom.servletrunner.util.*;
/**
 * created by the runner. spawns servlet instances
 *
 * 08.04.02 ho: queued_startup
 * 19.03.02 ho: creation
 */
public class ServletContainer extends Application implements Runnable {
	// the configuration for this container. this object must be (at least) visible for the runner to get statistics
	public ServletContainerConfig scc = null;
	// the vector with the spawned instances
	private Vector vecInstances = null;
	// the watchdog thread for this container (checks if instances are running)
	private Thread thrdWatchdog = null;
	// flag: watchdog running ?
	private boolean blnIsRunning = false;
/**
 * 
 */
public ServletContainer(ServletContainerConfig scc) {
	super();
	this.scc = scc; // the config
	init();
}
/**
 * disposes the container (stopping ServletInstances and cleaning up)
 */
public void dispose() {
	blnDisposed = true;

	/**
	 * stop the watchdog so it doesn't try to restart the stopped instances
	 */
	blnIsRunning = false;
	scc.objNotifyDeath.lo_notify();
	try {
		synchronized (thrdWatchdog) {
			thrdWatchdog.join(0);
		}
	} catch (InterruptedException e) {
	} catch (NullPointerException e) {
		// the thread may stops faster than we can call .join() -> ignore
	}

	/**
	 * dispose the instances (last created -> first stopped) and remove them from the vector
	 */
	if (vecInstances != null) {
		for (int i = vecInstances.size(); i > 0; i--) {
			scc.logContext.write("D Disposing instance " + i + "/" + vecInstances.size() + " for " + scc.strProcessName);
			try {
				((ServletInstance) vecInstances.elementAt(i - 1)).dispose();
			} catch (Throwable t) {
				scc.logContext.write("? Unhandled exception while disposing instance", t);
			}
		}
		vecInstances.removeAllElements();
		vecInstances = null;
	}

	/**
	 * dispose the container configuration
	 */
	if (scc != null) {
		scc.dispose();
		scc = null;
	}

	/**
	 * we're done
	 */
}
/**
 * generates a servet-id in the form of "processname [nb/maxnb]"
 */
private String getServletID(int intThrdNb) {
	StringBuffer strb = new StringBuffer(20);
	strb.append(scc.strProcessName);
	strb.append("[");
	strb.append(intThrdNb + 1);
	strb.append("/");
	strb.append(scc.intNbOfInstances);
	strb.append("]");
	return strb.toString();
}
/**
 * initialisation and spawning the Servlet Instances
 */
private void init() {
	// vector with the ServletInstances
	vecInstances = new Vector(scc.intNbOfInstances);
	// create a threadgroup for the instances 
	scc.thrdgInstances = new ThreadGroup(scc.strServletIdentifier);
	// create the instances. kill the container if one of the instances didn't start.
	scc.logContext.write("# Creating " + scc.intNbOfInstances + " instances for " + scc.strProcessName);
	for (int i = 0; i < scc.intNbOfInstances; i++) {
		String strID = getServletID(i);
		scc.logContext.write("D Starting instance " + strID);
		try {
			ServletInstance si = startInstance(strID);
			// we got NULL if something went wrong during init.
			if (si == null) {
				dispose();
				return;
			}
			vecInstances.addElement(si);
			scc.logContext.write("D Instance " + strID + " started");
		} catch (Throwable t) {
			scc.logContext.write("? Unhandled exception while creating instance", t);
			dispose();
			return;
		}
	}
	scc.logContext.write("# " + vecInstances.size() + " instances created for " + scc.strProcessName);
	// done. now start the watchdog for dead ServletInstances
	scc.logContext.write("# Trying to start watchdog thread for " + scc.strProcessName);
	blnIsRunning = true;
	thrdWatchdog = new Thread(scc.thrdgInstances, this, "Watchdog for " + scc.strProcessName);
	thrdWatchdog.setPriority(Thread.MIN_PRIORITY);
	thrdWatchdog.start();
}
/**
 * watchdog-thread. is notified by an instance and tries to restart the instance again.
 * if the instance crashes in the .init() now, we blank-out the instance in the
 * vector and try to .dispose() the session.
 *
 * we also look after every 10 seconds for dead instances
 *
 * XXX ho: 
 */
public void run() {
	//
	scc.logContext.write("# Thread started");
	//
	while (blnIsRunning) {
		/**
		 * the scc.objNotifyDeath notified by a ServletInstance if something went wrong.
		 */
		scc.objNotifyDeath.lo_wait(10l * 1000l); // 10 seconds
		// shall we stop working ? (.dispose())
		if (!blnIsRunning) {
			scc.logContext.write("# Thread stopped");
			break;
		}

		/**
		 * now crawl through the stored instances and check the status of every instance
		 */
		for (int i = 0; i < vecInstances.size(); i++) {
			ServletInstance si = (ServletInstance) vecInstances.elementAt(i);
			// if the servletinstance crashed during restart, we have NULL here and we ignore the instance
			if (si == null)
				continue;
			//
			String strID = getServletID(i); // build the process-id (for logging)
			boolean blnInstanceOK = si.isRunning(); // get the state of the process
			// dead thread ? try to recreate it
			if (!blnInstanceOK) {
				// first try destroy the instance to kill the MonitoredExecution thread
				scc.logContext.write("! Destroying instance " + strID);
				try {
					si.dispose();
				} catch (Throwable t1) {
					scc.logContext.write("? Unhandled exception while destroying instance", t1);
				}
				scc.logContext.write("! Try to restart instance " + strID);
				try {
					si = startInstance(strID);
					// if the instance is running again now, we're happy.
					if (si != null && si.isRunning()) {
						blnInstanceOK = true;
						scc.stats.intInstancesRestarted++;
						scc.logContext.write("! Instance " + strID + " recreated ");
						vecInstances.setElementAt(si, i); // isn't really needed, only for clarification
					}
				} catch (Throwable t1) {
					// there was an error while starting the instance ?
					scc.logContext.write("? Unhandled exception while creating instance", t1);
				}
				// lets check again, if we were able to create the instance.
				if (!blnInstanceOK) {
					try {
						si.dispose();
					} catch (Throwable t) {
					}
					vecInstances.setElementAt(null, i);
					scc.logContext.write("? Cannot recreate intance " + strID);
				}
			}
		}
	}
	blnIsRunning = false;
}
/**
 * tries to (re)start a servlet instance
 */
private ServletInstance startInstance(String strID) throws Throwable {
	/**
	 * this line is left as an exercise to the reader :-)
	 */
	ServletInstance si = new ServletInstance(scc, strID);

	/**
	 * just to be sure.
	 */
	if (si == null)
		return null;

	/**
	 * .isRunning() is true when we are at the end of .init() and returning from the constructor.
	 * otherwise we had an error in ServletInstance which was written into the logfile.
	 */
	if (!si.isRunning())
		return null;

	//
	return si;
}
}