package net.conacom.servletrunner.sessionmanager;

/**
 * Copyright 1999-2002 by conacom GmbH
 * Wilhelmshoeher Strasse 202, 60389 Frankfurt/Main, Germany
 * All rights reserved.
 * 
 * This software or implementation is free for commercial and non-commercial use.
 * No warranty is given about the functionality.
 */
import java.io.*;
import java.util.*;
import javax.servlet.http.*;
import net.conacom.servletrunner.util.*;
import net.conacom.servletrunner.impl.*;
import net.conacom.servletrunner.main.*;
/**
 * a sessionmanger who stores basic data in memory but serializes the hashtable with the values to disk
 *
 * configdata: directory=directory to the sessiondata
 *
 * 11.05.02 ho: fixed bug in a errormessage
 * 08.04.02 ho: sessioncounter isnt static anymore
 * 24.03.02 ho: creation
 */
public class FileSessionManager extends SessionManager {
	private Hashtable hshHeaders = null;
	private String strPath = null;
	private int intMaxFileRemovePerScan = 0;
	private int intSessionCount = 0;
	// the header
	private class Header {
		String strID;
		long lngCreationTime = -1;
		long lngLastAccessedTime = -1;
		long lngMaxInactiveInterval = -1;
		public Header(SessionData sd) {
			strID = sd.strID;
			lngCreationTime = sd.lngCreationTime;
			lngLastAccessedTime = sd.lngLastAccessedTime;
			lngMaxInactiveInterval = sd.lngMaxInactiveInterval;
		}
		public SessionData getSessionData() {
			SessionData sd = new SessionData();
			sd.strID = strID;
			sd.lngCreationTime = lngCreationTime;
			sd.lngLastAccessedTime = lngLastAccessedTime;
			sd.lngMaxInactiveInterval = lngMaxInactiveInterval;
			sd.hshValues = null;
			return sd;
		}
	}
/**
 * MemorySessionStorage - Konstruktorkommentar.
 * @param log net.conacom.servletrunner.main.LogContext
 * @param tfTime net.conacom.servletrunner.util.TimeFactory
 * @param lngSessionTimeout long
 * @param lngCleanupIdleTime long
 */
public FileSessionManager(LogContext log, Properties prp) {
	super(log, prp);
	init();
}
/**
 * cleanup of outdated sessions
 */
protected void cleanup() {
	synchronized (getLock()) {
		log.write("D Starting cleanup");
		int intCntCleaned = 0;
		int intCntTotal = 0;
		for (Enumeration e = hshHeaders.elements(); e.hasMoreElements();) {
			intCntTotal++;
			Header h = (Header) e.nextElement();
			SessionData sd = h.getSessionData();
			if (!sd.isValidSession()) {
				hshHeaders.remove(sd.strID);
				File f = new File(strPath, h.strID);
				if (!f.delete()) {
					log.write("? Cannot delete " + f.getPath());
				}
				intCntCleaned++;
				if (intCntCleaned > intMaxFileRemovePerScan) {
					log.write("! Limit of files to clean reached, will continue in the next cycle");
					break;
				}
			}
		}
		log.write("# Removed " + intCntCleaned + " entries of " + intCntTotal);
	}
}
/**
 * 
 */
public SessionData createSession(HttpServletRequest req) {
	long lngTime = TimeFactory.getTime();
	//
	SessionData sc = new SessionData();
	sc.blnIsNewSession = true;
	sc.hshValues = new Hashtable();
	sc.lngCreationTime = lngTime;
	sc.lngLastAccessedTime = lngTime;
	sc.lngMaxInactiveInterval = lngSessionTimeout;
	// maybe we shall encrypt the IP also into the ID
	sc.strID = "ID" + Long.toHexString(lngTime) + "-" + Integer.toString(++intSessionCount);
	//
	return sc;
}
/**
 * gets a session by using ObjectInputStream
 */
protected SessionData getSession(String strID) {
	synchronized (getLock()) {
		Header h = (Header) hshHeaders.get(strID);
		SessionData sd = h.getSessionData();
		File f = new File(strPath, h.strID);
		if (!f.exists())
			return null;
		if (sd.hshValues != null) {
			sd.hshValues.clear();
			sd.hshValues = null;
		}
		ObjectInputStream ois = null;
		try {
			ois = new ObjectInputStream(new FileInputStream(f));
			sd.hshValues = (Hashtable) ois.readObject();
		} catch (IOException e) {
			log.write("? Error reading sessisondata.", e);
			return null;
		} catch (ClassNotFoundException e) {
			log.write("? Error reading sessisondata.", e);
			return null;
		} catch (Throwable e) {
			log.write("? Unknown error while reading sessisondata", e);
			return null;
		}
		Helper.close(ois);
		// just to be save
		if (sd.hshValues == null)
			sd.hshValues = new Hashtable();
		return sd;
	}
}
/**
 * 
 * @return java.lang.String
 */
public String getStatistics() {
	synchronized (getLock()) {
		StringBuffer strb = new StringBuffer();
		strb.append("Entries total: ");
		strb.append(hshHeaders.size());
		strb.append(", valid entries: ");
		int intValid = 0;
		for (Enumeration e = hshHeaders.elements(); e.hasMoreElements();) {
			Header h = (Header) e.nextElement();
			SessionData sd = h.getSessionData();
			if (sd.hshValues != null && sd.isValidSession()) {
				intValid++;
			}
		}
		strb.append(intValid);
		return strb.toString();
	}
}
/**
 * 
 */
private void init() {
	hshHeaders = new Hashtable();
	strPath = prpConfig.getProperty("directory");
	try {
		intMaxFileRemovePerScan = Integer.parseInt(prpConfig.getProperty("max_file_remove_per_scan"));
	} catch (NumberFormatException e) {
		log.write("! Cannot parse your 'max_file_remove_per_scan' parameter, using default");
		intMaxFileRemovePerScan = 75;
	}
}
/**
 * session for ID is valid ?
 */
public boolean isValid(String strID) {
	if (strID == null)
		return false;
	//
	synchronized (getLock()) {
		// do we have a header for this ID ?
		Header h = (Header) hshHeaders.get(strID);
		if (h == null)
			return false;
		// do we have the file for this ID ?
		File f = new File(strPath, h.strID);
		if (!f.exists()) {
			log.write("? Have header but no file for it !");
			return false;
		}
		// get the sessiondata from the header and validate it
		SessionData sd = h.getSessionData();
		return sd.isValidSession();
	}
}
/**
 * save session using ObjectOutputStream
 */
protected void putSession(SessionData sd) {
	synchronized (getLock()) {
		Header h = new Header(sd);
		File f = new File(strPath, h.strID);
		ObjectOutputStream oos = null;
		try {
			oos = new ObjectOutputStream(new FileOutputStream(f));
			oos.writeObject(sd.hshValues);
		} catch (IOException e) {
			log.write("? Error writing sessisondata", e);
		}
		Helper.flush(oos);
		Helper.close(oos);
		//
		sd.hshValues.clear();
		sd.hshValues = null;
		hshHeaders.put(h.strID, h);
	}
}
}