package net.conacom.servletrunner.sessionmanager;

/**
 * Copyright 1999-2002 by conacom GmbH
 * Wilhelmshoeher Strasse 202, 60389 Frankfurt/Main, Germany
 * All rights reserved.
 * 
 * This software or implementation is free for commercial and non-commercial use.
 * No warranty is given about the functionality.
 */
import java.util.*;
import javax.servlet.http.*;
import net.conacom.servletrunner.util.*;
import net.conacom.servletrunner.main.*;
import net.conacom.servletrunner.impl.*;
/**
 * manages sessions. outputs into the runner-logfile because we have only
 * one instance for all containers.
 *
 * 25.06.02 ho: removed getDefaultSessionTimeout(). session-id is configurable
 * 23.05.02 ho: getDefaultSessionTimeout();
 * 21.04.02 ho: creation
 */
public abstract class SessionManager extends Application {
	public final static String SM_SESSIONTIMEOUT = "sessiontimeout";
	public final static String SM_CLEANUPCYCLE = "sessioncleanupcycle";
	public final static String SM_SESSION_ID = "session_id";
	//
	protected Properties prpConfig = null;
	protected LogContext log = null;
	// should be private, but maybe sb. other need this
	protected long lngSessionTimeout = 0;
	protected long lngCleanupCycle = 0;
	//
	private Object objLock = new Object();
	private CleanupThread thrdCleaner = null;

	/**
	 * local thread: calls cleanup after lngCleanupCycle
	 * PRIORITY = MAX !
	 */
	private class CleanupThread extends Thread {
		private boolean blnIsRunning = true;
		private long lngWaitTime = 0;
		private CleanupThread(long l) {
			lngWaitTime = l;
			blnIsRunning = true;
			setName("SessionManager-Cleanup");
			setPriority(Thread.MAX_PRIORITY);
			start();
		}
		private void dispose() {
			log.write("D Disposing SessionManager-Cleanup thread");
			blnIsRunning = false;
			synchronized (this) {
				notify();
				try {
					join();
				} catch (InterruptedException e) {
				}
			}
		}
		public void run() {
			while (blnIsRunning) {
				synchronized (this) {
					try {
						wait(lngWaitTime);
					} catch (InterruptedException e) {
					}
				}
				if (!blnIsRunning)
					break;
				//
				cleanup();
			}
		}
	}
/**
 * SessionStorageObject - Konstruktorkommentar.
 */
public SessionManager(LogContext log, Properties prpConfig) {
	super();
	this.log = log;
	this.prpConfig = prpConfig;
	init0();
}
/**
 * cleanup outdated Sessions.
 * this process must synchronize his access to the storage with .getLock()
 * to avoid concurrent reading
 */
protected abstract void cleanup();
/**
 * creates a new HttpSession. the HttpServletRequest is for grabbing informations
 * about the client-ip and others which may come in use if we support
 * IDs in the URL
 */
public HttpSession_Impl createHttpSession(HttpServletRequest req) {
	return new HttpSession_Impl(createSession(req));
}
/**
 * creates a new HttpSession. the HttpServletRequest is for grabbing informations
 * about the client-ip and others which may come in use if we support
 * IDs in the URL
 */
public abstract SessionData createSession(HttpServletRequest req);
/**
 * 
 */
public void dispose() {
	// avoid multiple-calls
	if (blnDisposed)
		return;
	blnDisposed = true;
	//
	if (thrdCleaner != null) {
		thrdCleaner.dispose();
	}
	cleanup();
	//
	objLock = null;
}
/**
 * forces the cleanup. does nothing if we're currently cleaning
 */
public void forceCleanup() {
	synchronized (thrdCleaner) {
		thrdCleaner.notify();
	}
}
/**
 * gets a HttpSession session by its ID
 * returns NULL if the session cannot be found or is invalid
 */
public HttpSession_Impl getHttpSession(String strID) {
	if (!isValid(strID))
		return null;
	//
	SessionData sc = getSession(strID);
	if (sc == null)
		return null;
	//
	sc.blnIsNewSession = false;
	sc.lngLastAccessedTime = TimeFactory.getTime();
	return new HttpSession_Impl(sc);
}
/**
 * central lock object for synchronisation
 */
protected Object getLock() {
	return objLock;
}
/**
 * retrieves sessionconfig by ID
 *.isValid(...) must be called before this method so we can be sure that the session exists and is valid
 *
 * returns NULL if something went wrong
 */
protected abstract SessionData getSession(String strID);
/**
 * 
 * @return java.lang.String
 */
public abstract String getStatistics();
/**
 * 
 */
private void init0() {
	//
	String s = null;
	try {
		s = prpConfig.getProperty(SM_CLEANUPCYCLE);
		this.lngCleanupCycle = (long) (Double.valueOf(s).doubleValue() * 60d * 1000d);
	} catch (NumberFormatException e) {
		log.write("? Sessionmanager: Error parsing value '" + s + "' for parameter '" + SM_CLEANUPCYCLE + "'");
	}
	//
	try {
		s = prpConfig.getProperty(SM_SESSIONTIMEOUT);
		this.lngSessionTimeout = (long) (Double.valueOf(s).doubleValue() * 60d * 1000d);
	} catch (NumberFormatException e) {
		log.write("? Sessionmanager: Error parsing value '" + s + "' for parameter '" + SM_CLEANUPCYCLE + "'");
	}
	// session id
	s = prpConfig.getProperty(SM_SESSION_ID);
	if (s != null)
		HttpSession_Impl.SESSION_IDENTIFIER = s;
	//
	log.write("# Starting session-cleaner");
	thrdCleaner = new CleanupThread(lngCleanupCycle);
	if (thrdCleaner != null)
		log.write("# Session-cleaner started");
}
/**
 * checks if a stored session is valid or timed-out
 *
 * returns FALSE if the session is timed-out or wasn't found or the ID is NULL
 *
 * we need this method to find out if a session is outdated w/o
 * retrieving the whole of data (needed by cleanup(), getHttpSession() and
 * the HttpServletRequest_Impl too
 */
public abstract boolean isValid(String strID);
/**
 * external access for storing the session
 */
public void putHttpSession(HttpSession_Impl hs) {
	SessionData sd = hs._getSessionData();
	putSession(sd);
}
/**
 * must write the session asap to memory/disk/database ...
 */
protected abstract void putSession(SessionData sd);
}