package net.conacom.servletrunner.util;

/**
 * Copyright 1999-2002 by conacom GmbH
 * Wilhelmshoeher Strasse 202, 60389 Frankfurt/Main, Germany
 * All rights reserved.
 * 
 * This software or implementation is free for commercial and non-commercial use.
 * No warranty is given about the functionality.
 */
import java.util.Enumeration;
import java.util.Locale;
import java.awt.FontMetrics;
import java.io.Serializable;
import java.util.Dictionary;
/**
* ByteString maps String functionality to a byte[] for a better support of byte streams that contain
* text data.
* @author: Karl Schmidt, <a href="mailto:schmidt@conacom.de">Karl_Schmidt@bigfoot.com</a>
* @version 1.0 Copyright (C) 2002 by Karl Schmidt - all rights reserved
*/
public final class ByteString implements Serializable, Cloneable {
	// conversion constants
	private static final int CASEDIFF = ('a' - 'A');
	private static final int LOWER_BIT = 32;
	private static final int UPPER_BIT = (byte) 0xdf;
	private static final byte[] NULL = "null".getBytes();
	private static final byte[] TRUE = "true".getBytes();
	private static final byte[] FALSE = "false".getBytes();

	// number conversion helper
	private final static byte[] DIGITS = "0123456789".getBytes();
	private final static byte[] UNITS = "0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789".getBytes();
	private final static byte[] TENTH_UNITS = "0000000000111111111122222222223333333333444444444455555555556666666666777777777788888888889999999999".getBytes();
	private final static byte[] INTEGER_MIN = "-2147483648".getBytes();
	private final static byte[] LONG_MIN = "-9223372036854775808".getBytes();
	private final static byte MINUS = 45;

	// the char values
	private byte[] bytValue = null;
	// the total length
	private int intLength = 0;

	// next line position
	private int intNextLineStart = 0;

	// the line seperator for this object
	private byte[] bytLineBreak = {13, 10};

	// enumerates through lines
	private class LineEnumeration implements Enumeration {
		private int intLineCounter = 0;
		private int intType = 0;
		private int intWidth = 0;
		private FontMetrics fmMetrics = null;
		LineEnumeration() {
		}
		LineEnumeration(int width) {
			intType = 1;
			intWidth = width;
		}
		LineEnumeration(int width, FontMetrics fm, boolean skipSpaces) {
			if (skipSpaces) {
				intType = 2;
			} else {
				intType = 3;
			}
			intWidth = width;
			fmMetrics = fm;
		}
		public Object nextElement() {
			ByteString line;
			switch (intType) {
				case 0 :
					line = getLineHardBreak(intLineCounter);
					break;
				case 1 :
					// line = getLineFixedLength(intLineCounter, intWidth);
					break;
				case 2 :
					// line = getLineSoftLength(intLineCounter, intWidth, fmMetrics, true);
					break;
				case 3 :
					// line = getLineSoftLength(intLineCounter, intWidth, fmMetrics, false);
					break;
				default :
					line = new ByteString();
			}
			intLineCounter = intNextLineStart;
			return line;
		}
		public boolean hasMoreElements() {
			return intLineCounter < intLength;
		}
	}
/**
* Constructs an empty ByteString.
*/
public ByteString() {
	this(16);
}
/**
* Constructs an empty ByteString object with an initial capacity specified by the 
* <code>length</code>&nbsp;argument.
* @param length the initial capacity.
*/
public ByteString(int length) {
	bytValue = new byte[length];
}
/**
* Constructs a ByteString object that represents the same sequence of characters as the string argument.
* @param str the initial contents of the buffer.
*/
public ByteString(String str) {
	this(str.length() + 16);
	append(str);
}
/**
* Constructs a ByteString object that represents the same sequence of characters as the ByteString argument.
* @param str the initial contents of the buffer.
*/
public ByteString(ByteString str) {
	this(str.bytValue, 0, str.intLength);
}
/**
* Constructs a ByteString object that represents the same sequence of characters as the ByteString argument.
* @param str the initial contents of the buffer.
* @param off the first data char copied to the new ByteString
* @param len the total amount of chars copied into the new ByteString
*/
public ByteString(ByteString str, int off, int len) {
	this(str.bytValue, off, len);
}
/**
* Constructs a ByteString object that represents the same sequence of characters as the 
* byte array argument.
* @param chars the initial contents of the buffer.
*/
public ByteString(byte[] chars) {
	this(chars, 0, chars.length);
}
/**
* Constructs a ByteString that represents the same sequence of characters as the byte array
* argument, beginning at 'off', using 'len' characters.
* @param chars the initial contents of the buffer.
* @param off the first data char copied to the ByteString
* @param len the total amount of chars copied to the ByteString
*/
public ByteString(byte[] chars, int off, int len) {
	this(chars.length + 16);
	append(chars, off, len);
}
/**
* Constructs a ByteString object that represents the same sequence of characters as the char array argument.
* @param chars the initial contents of the buffer.
*/
public ByteString(char[] chars) {
	this(chars, 0, chars.length);
}
/**
* Constructs a ByteString that represents the same sequence of characters as the char array argument,
* beginning at 'off', using 'len' characters.
* @param chars the initial contents of the buffer.
* @param off the first data char copied to the ByteString
* @param len the total amount of chars copied to the ByteString
*/
public ByteString(char[] chars, int off, int len) {
	this(chars.length + 16);
	append(chars, off, len);
}
/**
* Appends the <code>byte</code>&nbsp;argument to this ByteString.
* @param b a <code>byte</code>.
* @return returns this ByteString.
*/
public final ByteString append(byte b) {
	int newcount = intLength + 1;
	if (newcount > bytValue.length)
		ensureCapacity(newcount);
	bytValue[intLength++] = b;
	return this;
}
/**
* Appends the <code>char</code>&nbsp;argument to this ByteString.
* @param ch a <code>char</code>.
* @return returns this ByteString.
*/
public final ByteString append(char c) {
	int newcount = intLength + 1;
	if (newcount > bytValue.length)
		ensureCapacity(newcount);
	bytValue[intLength++] = (byte) c;
	return this;
}
/**
* Appends the string representation of the <code>double</code>&nbsp;argument to this ByteString. 
* @param d a <code>double</code>.
* @return returns this ByteString.
*/
public final ByteString append(double d) {
	return append(String.valueOf(d));
}
/**
* Appends the string representation of the <code>float</code>&nbsp;argument to this ByteString. 
* @param f a <code>float</code>.
* @return returns this ByteString.
*/
public final ByteString append(float f) {
	return append(String.valueOf(f));
}
public final ByteString append(int i) {
	// The calculation is done in the positive and Integer.MIN_VALUE has no positive expression
	if (i == Integer.MIN_VALUE) {
		return append(INTEGER_MIN, 0, INTEGER_MIN.length);
	}
	byte[] buf = new byte[12];
	boolean negative = (i < 0);
	int charPos = 12;
	if (negative) {
		i = -i;
		// append minus sign
		append(MINUS);
	}
	if (i < 10) {
		append((byte) (i + 48));
		return this;
	}
	byte[] units = UNITS;
	byte[] tenthUnits = TENTH_UNITS;
	do {
		int digit = i % 100;
		buf[--charPos] = units[digit];
		buf[--charPos] = tenthUnits[digit];
		i = i / 100;
	} while (i != 0);
	// if the last 100 block had only one digit, go one back
	if (buf[charPos] == 48) {
		charPos++;
	}
	append(buf, charPos, (12 - charPos));
	return this;
}
public final ByteString append(long l) {
	// The calculation is done in the positive and Long.MIN_VALUE has no positive expression
	if (l == Long.MIN_VALUE) {
		return append(LONG_MIN, 0, LONG_MIN.length);
	}
	byte[] buf = new byte[20];
	boolean negative = (l < 0);
	int charPos = 20;
	if (negative) {
		l = -l;
		// append minus sign
		append(MINUS);
	}
	if (l < 10) {
		return append((byte) (l + 48));
	}
	byte[] units = UNITS;
	byte[] tenth = TENTH_UNITS;
	do {
		int digit = (int) (l % 100);
		buf[--charPos] = units[digit];
		buf[--charPos] = tenth[digit];
		l = l / 100;
	} while (l != 0);
	// if the last 100 block had only one digit, go one back
	if (buf[charPos] == 48) {
		charPos++;
	}
	return append(buf, charPos, (20 - charPos));
}
/**
* Appends the string representation of the <code>Object</code>&nbsp;argument to this StringBufferExtended. 
* @param obj an <code>Object</code>.
* @return returns this StringBufferExtended.
*/
public final ByteString append(Object obj) {
	return append(String.valueOf(obj));
}
/**
* Appends the string to this ByteString. 
* @param str a string.
* @return returns this ByteString.
*/
public final ByteString append(String str) {
	if (str == null) {
		return append(NULL, 0, 4);
	}
	int len = str.length();
	int newcount = intLength + len;
	if (newcount > bytValue.length)
		ensureCapacity(newcount);
	byte[] b = getASCII(str);
	System.arraycopy(b, 0, bytValue, intLength, len);
	intLength = newcount;
	return this;
}
/**
* Appends the string to this ByteString.
* @param str a string.
* @return returns this ByteString.
*/
public final ByteString append(ByteString str) {
	if (str == null) {
		return append(NULL, 0, 4);
	}
	int len = str.intLength;
	int newcount = intLength + len;
	if (newcount > bytValue.length)
		ensureCapacity(newcount);
	System.arraycopy(str.bytValue, 0, bytValue, intLength, len);
	intLength = newcount;
	return this;
}
/**
* Appends the string representation of the <code>boolean</code>&nbsp;argument to the ByteString. 
* @param b a <code>boolean</code>.
* @return returns this ByteString.
*/
public final ByteString append(boolean b) {
	if (b)
		return append(TRUE, 0, 4);
	return append(FALSE, 0, 5);
}
/**
* Appends the ASCII representation of the bytes from the <code>byte</code>&nbsp;array argument to this
* ByteString.
* @param b the bytes to be appended.
* @return returns this ByteString.
*/
public final ByteString append(byte[] b) {
	return append(b, 0, b.length);
}
/**
* Appends the ASCII representation of the bytes from the <code>byte</code>&nbsp;array argument to this
* ByteString. 
* @param b the bytes to be appended.
* @param offset the index of the first character to append.
* @param len the number of characters to append.
* @return returns this ByteString.
*/
public final ByteString append(byte[] b, int offset, int len) {
	int newcount = intLength + len;
	if (newcount > bytValue.length)
		ensureCapacity(newcount);
	System.arraycopy(b, offset, bytValue, intLength, len);
	intLength = newcount;
	return this;
}
/**
* Appends all chars from the <code>char</code>&nbsp;array argument to this ByteString.
* @param chars the characters to be appended.
* @return returns this ByteString.
*/
public final ByteString append(char[] chars) {
	return append(chars, 0, chars.length);
}
/**
* Appends some chars from the <code>char</code>&nbsp;array argument to this ByteString. 
* @param chars the characters to be appended.
* @param offset the index of the first character to append.
* @param len the number of characters to append.
* @return returns this ByteString.
*/
public final ByteString append(char[] chars, int offset, int len) {
	int newcount = intLength + len;
	if (newcount > bytValue.length)
		ensureCapacity(newcount);
	System.arraycopy(getASCII(chars), offset, bytValue, intLength, len);
	intLength = newcount;
	return this;
}
/**
* appends a line break at the end of this string
* @return returns this ByteString
*/
public ByteString appendLineBreak() {
	append(bytLineBreak, 0, bytLineBreak.length);
	return this;
}
/**
* sets size of this ByteString to <code>0</code>.  The content or the size of the buffer is not
* changed.
*/
public final void clear() {
	intLength = 0;
}
/**
* returns a clone of this ByteString.
* @return returns returns a clone of this ByteString.
*/
public final Object clone() {
	try {
		ByteString sbe = (ByteString) super.clone();
		sbe.bytValue = (byte[]) bytValue.clone();
		return sbe;
	} catch (CloneNotSupportedException e) {
		throw new InternalError("clone failed on StringBufferExtended");
	}
}
/**
* Compares this ByteString to the <code>ByteString</code>&nbsp;argument.
* @return returns the value <code>0</code>&nbsp;if the argument is equal to this ByteString; a
* value less than <code>0</code>&nbsp;if this ByteString is lexicographically less than the
* argument; and a value greater than <code>0</code>&nbsp;if this ByteString is lexicographically
* greater than the argument.
* @param comp a ByteString
* @return returns an int value indicating the difference
* @see #compareTo(char[], int, int)
*/
public final int compareTo(ByteString comp) {
	return compareTo(comp.bytValue, 0, comp.intLength, false);
}
/**
* Compares the content of a byte array lexicographically to this ByteString. The comparison is
* based on the byte value of each character.
* <P>The byte array is considered equal to the ByteString, if it is of the same length and
* contains the same sequence of characters.  It is considered lexicographically less, if at least one
* character has a byte value smaller than the character at the corresponding position in this
* ByteString, or if it is shorter than this ByteString.  It is considered greater, if
* at least one character has a byte value greater than the character at the corresponding position in
* this ByteString, or if it is longer than this ByteString.
* <P>This method allows the comparison of a portion of the byte array argument.  It will compare
* <code>len</code> characters of the byte array argument, starting at position <code>off</code>, to the
* characters in this ByteString.  The comparison stops if either <code>len</code> characters have
* been compared or the end of the ByteString is reached.
* <P>Since the comparison is based on the byte values of the characters, the result does not always
* represent a real lexicographically order for all languages.  E.g. an '&auml;' character is considered to
* be greater than a 'z'.  When using the parameter <code>ignoreCase</code> the matter grows even worse.
* The comparison will ignore any characters which do not represent an ASCII letter.
*
* @param comp byte array to be compared.
* @param off the first character of the argument to compare
* @param len the total number of chars to compare
* @param ignoreCase if true, the case is ignored.
*
* @return the value <code>0</code> if the argument characters in the give range are equal to the characters
* in this ByteString; a value less than <code>0</code> if the characters in this
* ByteString are lexicographically less than the characters in the argument array; and a value
* greater than <code>0</code> if the characters in this ByteString are lexicographically greater
* than the characters in the argument.
*/
public final int compareTo(byte[] chars, int off, int len, boolean ignoreCase) {
	if (off < 0 || len < 0 || len + off > chars.length)
		throw new IllegalArgumentException();

	//
	int shorter = Math.min(intLength, len);
	byte[] source = bytValue;
	int sourceIndex = 0;
	int argIndex = off;
	while (shorter-- != 0) {
		int sourceChar = source[sourceIndex++] & 0x000000ff;
		int argChar = chars[argIndex++] & 0x000000ff;
		if (sourceChar == argChar)
			continue;
		if (ignoreCase) {
			int upper = sourceChar & UPPER_BIT;
			// only compare 'real' letters A-Z (65-90)
			if (upper > 64 && upper < 91 && upper == (argChar & UPPER_BIT))
				continue;
		}
		return sourceChar - argChar;
	}
	return intLength - len;
}
/**
* Compares the <code>ByteString</code>&nbsp;argument lexicographically to this ByteString.
* See <code>compareToIgnoreCase(char[], int, int)</code>&nbsp; for a more detailed description and some of
* the problems.
*
* @param comp the <code>String</code>&nbsp;to be compared.
*
* @return returns the value <code>0</code>&nbsp;if the argument is equal to this ByteString; a
* value less than <code>0</code>&nbsp;if this ByteString is lexicographically less than the
* argument; and a value greater than <code>0</code>&nbsp;if this ByteString is lexicographically
* greater than the argument.
* @see #compareToIgnoreCase(char[], int, int)
*/
public final int compareToIgnoreCase(ByteString comp) {
	return compareTo(comp.bytValue, 0, comp.intLength, true);
}
/**
* Converts all of the characters in this string buffer to lower case which are in the range of 'A'-'Z'.
* When working with technical strings like when parsing html or a protocol like SMTP or POP it is quite
* inconvenient when letters suddenly convert to something strange because the computer uses some exotic
* locale.
* @return returns the ByteString, converted all ASCII characters to lowercase.
*/
public final ByteString convertToLowerCase() {
	int len = intLength;
	byte[] val = bytValue;
	//	
	for (int i = 0; i < len; ++i) {
		int ch = val[i] & 0x000000ff;
		if (ch > 64 && ch < 91) {
			val[i] |= LOWER_BIT;
		}
	}
	return this;
}
/**
* Converts all of the characters in this string buffer to lower case which are in the range of 'a'-'z'.
* When working with technical strings like when parsing html or a protocol like SMTP or POP it is quite
* inconvenient when letters suddenly convert to something strange because the computer uses some exotic
* locale.
* @return returns the ByteString, converted all ASCII characters to lowercase.
*/
public final ByteString convertToUpperCase() {
	int len = intLength;
	byte[] val = bytValue;

	//	
	for (int i = 0; i < len; ++i) {
		int ch = val[i] & 0x000000ff;
		if (ch > 96 && ch < 123) {
			val[i] &= CASEDIFF;
		}
	}
	return this;
}
/**
* decrements the length of the ByteString by one.  This is a convenience method to delete single
* characters at the end of a ByteString.
* @return returns this ByteString
*/
public final ByteString decLength() {
	if (--intLength < 0)
		intLength = 0;
	return this;
}
/**
* deletes all chars from the end of the ByteString, beginning at position <code>off</code>.
* @param off the first char to delete, must be >= 0 and < length.
* @return returns this ByteString
*/
public final ByteString delete(int off) {
	return delete(off, Integer.MAX_VALUE);
}
/**
* deletes <code>len</code>&nbsp; chars of the ByteString, beginning at position <code>off</code>.
* @param off the first char to delete, must be &gt;= 0 and &lt; length.
* @param len the total amount of chars to delete, may be any positive number, or 0.  If the value of the len
* argument plus the value of the offset argument is greater than the length of this buffer, the end of the
* buffer, beginning at the position <code>off</code>&nbsp;, will be deleted.
* @return returns this ByteString
*/
public final ByteString delete(int off, int len) {
	if (off < 0 || off >= intLength)
		throw new StringIndexOutOfBoundsException("offset out of bounds: " + off);
	if (len < 0)
		throw new StringIndexOutOfBoundsException("len must be >= 0: " + len);

	// if len == null do nothing
	if (len == 0)
		return this;
	int overflowCheck = off + len;
	if (overflowCheck >= intLength || overflowCheck < 0) {
		setLength(off);
		return this;
	}

	// copy the array backwards
	System.arraycopy(bytValue, off + len, bytValue, off, intLength - (off + len));
	intLength -= len;
	return this;
}
/**
* releases all internal resources.
*/
public void dispose() {
	bytValue = null;
	intLength = 0;
	intNextLineStart = 0;
}
/**
* Tests if this ByteString ends with the specified suffix.
*
* @param str a String containing the suffix.
*
* @return <code>true</code>&nbsp; if the character sequence represented by the argument is a suffix of the
* character sequence represented by this object; <code>false</code>&nbsp; otherwise.
*/
public final boolean endsWith(String str) {
	int len = str.length();
	if (intLength < len)
		return false;
	return isMatchingRegion(getASCII(str), 0, len, intLength - len, false);
}
/**
* Tests if this ByteString ends with the specified suffix.
*
* @param str a String containing the suffix.
*
* @return <code>true</code>&nbsp; if the character sequence represented by the argument is a suffix of the
* character sequence represented by this object; <code>false</code>&nbsp; otherwise.
*/
public final boolean endsWith(ByteString str) {
	int len = str.intLength;
	if (intLength < len)
		return false;
	return isMatchingRegion(str.bytValue, 0, len, intLength - len, false);
}
/**
* Ensures that the capacity of the buffer is at least equal to the specified minimum.
* If the current capacity of this string buffer is less than the argument, then a new 
* internal buffer is allocated with greater capacity.
* @param minimumCapacity the minimum desired capacity.
*/
public final void ensureCapacity(int minimumCapacity) {
	if (minimumCapacity > bytValue.length) {
		int newCapacity = (bytValue.length + 1) * 2;
		if (minimumCapacity > newCapacity) {
			newCapacity = minimumCapacity;
		}
		byte[] newValue = new byte[newCapacity];
		System.arraycopy(bytValue, 0, newValue, 0, intLength);
		bytValue = newValue;
	}
}
/**
* Compares this ByteString to the specified object.
* The result is <code>true</code>&nbsp; if and only if the argument is not <code>null</code>&nbsp; and is an
* <code>ByteString</code>&nbsp; object that represents the same sequence of characters as this
* object. 
* @param o the object to compare this <code>ByteString</code>&nbsp; against.
* @return <code>true</code>&nbsp; if the objects are equal; <code>false</code>&nbsp; otherwise.
*/
public final boolean equals(Object o) {
	if (this == o) {
		return true;
	}
	if ((o != null) && (o.getClass() == getClass())) {
		ByteString anotherString = (ByteString) o;
		int n = intLength;
		if (n == anotherString.intLength) {
			int i = 0, j = n - 1;
			n = (n + 1) / 2;
			byte[] v1 = bytValue;
			byte[] v2 = anotherString.bytValue;
			while (n-- > 0) {
				if (v1[i] != v2[i++] || v1[j] != v2[j--]) {
					return false;
				}
			}
			return true;
		}
	}
	return false;
}
/**
* overwrites the characters from the given start position to the given end position, inclusive, with spaces.
* @param start the start position in the buffer.
* @param end the number of chars to overwrite.
* @return returns this ByteString.
*/
public final ByteString erase(int start, int end) {
	return overwrite(' ', start, end);
}
/**
* returns an array containing the lower byte of the unicode string characters
* @param s any string
* @return returns an array containing the lower byte of the unicode string characters
*/
public final static byte[] getASCII(String s) {
	char[] source = s.toCharArray();
	return getASCII(source, 0, source.length);
}
/**
* returns an array containing the lower byte of the unicode string characters
* @param s any string
* @return returns an array containing the lower byte of the unicode string characters
*/
public final static byte[] getASCII(char[] c) {
	return getASCII(c, 0, c.length);
}
/**
* returns an array containing the lower byte of the unicode string characters
* @param c any char array
* @param off the first char of the argument array to convert
* @param len the number of chars to be converted
* @return returns an array containing the lower byte of the unicode string characters
*/
public final static byte[] getASCII(char[] c, int off, int len) {
	byte[] result = new byte[len];
	for (int i = len; i-- > 0;) {
		result[i] = (byte) c[i + off];
	}
	return result;
}
/**
* returns the 'real' byte array from of this ByteString.
* @return returns the char array which is used for storing the ExtendedStringBuffers value internally.
*/
public final byte[] getByteArray() {
	return bytValue;
}
/**
* Returns the character at a specific index in this ByteString.
* <p>The index argument must be greater than or equal to <code>0</code>, and less than the length of this
* ByteString.
* @param index the index of the desired character.
* @return the character at the specified index of this ByteString.
*/
public final byte getByteAt(int index) {
	if (index >= intLength)
		throw new StringIndexOutOfBoundsException(index);
	return bytValue[index];
}
/**
* Characters are copied from this ByteString into the destination byte array <code>dst</code>. The
* first character to be copied is at index <code>begin</code>; the last character to be copied is at 
* index <code>end-1.</code> The total number of characters to be copied is <code>end-begin</code>. 
* The characters are copied into the subarray of <code>dst</code> starting at index <code>start</code>
* and ending at index:
* <p><blockquote><pre>
*     start + (end-begin) - 1
* </pre></blockquote>
* @param begin  start copying at this offset in the ByteString.
* @param end stop copying at this offset in the ByteString.
* @param dst the array to copy the data into.
* @param start offset into <code>dst</code>.
*/
public final void getBytes(int begin, int end, byte[] dst, int start) {
	if (begin < end) {
		System.arraycopy(bytValue, begin, dst, start, end - begin);
	}
}
/**
* Returns the current capacity of the ByteString. The capacity is the amount of storage available;
* beyond which an allocation will occur.
* @return returns the current capacity of this ByteString.
*/
public final int getCapacity() {
	return bytValue.length;
}
/**
* returns a char array containing the bytes of the byte array argument as low byte.  The high byte is 0.
* @param b a byte array
* @param off the index of the first byte in the argument array to be converted.
* @param len the total number of bytes to be converted.
* @return returns an array containing the char representation of the byte argument
*/
public final static char[] getCharArray(byte[] b, int off, int len) {
	char[] result = new char[len];
	for (int i = len; i-- > 0;) {
		result[i] = (char) (b[i + off] & 0xff);
	}
	return result;
}
/**
* Returns the character at a specific index in this ByteString.
* <p>The index argument must be greater than or equal to <code>0</code>, and less than the length of this
* ByteString.
* @param index the index of the desired character.
* @return the character at the specified index of this ByteString.
*/
public final char getCharAt(int index) {
	if (index >= intLength)
		throw new StringIndexOutOfBoundsException(index);
	return (char) (bytValue[index] & 0x000000ff);
}
/**
* returns the number of occurances of the specified character in this ByteString.
* @return returns the number of occurances of the specified character in this ByteString.
*/
public final int getCharCount(byte ch, int off, int len) {
	if (intLength == 0)
		return 0;
	byte[] value = bytValue;
	int count = 0;
	for (int i = intLength; i-- > 0;) {
		if (value[i] == ch)
			count++;
	}
	return count;
}
/**
* Returns the index within this ByteString of the first occurrence of the specified character.
* @param ch a character.
* @return the index of the first occurrence of the character in the character sequence represented by 
* this object, or <code>-1</code> if the character does not occur.
*/
public final int getIndexOf(byte ch) {
	return getIndexOf(ch, 0);
}
/**
* Returns the index within this ByteString of the first occurrence of the specified character,
* starting the search at the specified index.
* @param ch a character.
* @param fromIndex the index to start the search from.
* @return the index of the first occurrence of the character in the character sequence represented by this
* object that is greater than or equal to <code>fromIndex</code>, or <code>-1</code> if the character does
* not occur.
*/
public final int getIndexOf(byte ch, int fromIndex) {
	int max = intLength;
	byte[] v = bytValue;

	// check start index
	if (fromIndex < 0) {
		fromIndex = 0;
	} else {
		if (fromIndex >= intLength) {
			return -1;
		}
	}

	// loop through the chars
	for (int i = fromIndex; i < max; i++) {
		if (v[i] == ch) {
			return i;
		}
	}
	return -1;
}
/**
* Returns the index within this ByteString of the first occurrence of the specified substring.
*
* @param str a String containing the substring to search for.
*
* @return If the argument occurs as a substring within this object, then the index of the first character of the first such substring is 
* returned. If it does not occur as a substring, <code>-1</code> is returned.
*/
public final int getIndexOf(String str) {
	return getIndexOf(str, 0);
}
/**
* Returns the index within this ByteString of the first occurrence of the specified substring,
* starting at the specified index.
* @param str a String containing the substring to search for.
* @param fromIndex the index to start the search from.
* @return If the argument occurs as a substring within this object at a starting index no smaller 
* than <code>fromIndex</code>, then the index of the first character of the first such substring is 
* returned. If it does not occur as a substring starting at <code>fromIndex</code> or beyond, 
* <code>-1</code> is returned.
*/
public final int getIndexOf(String str, int from) {
	return getIndexOf(getASCII(str), 0, str.length(), from);
}
/**
* Returns the index within this ByteString of the first occurrence of the specified substring,
* starting at the specified index.
* @param str a ByteString containing the substring to search for.
* @param fromIndex the index to start the search from.
* @return If the argument occurs as a substring within this object at a starting index no smaller 
* than <code>fromIndex</code>, then the index of the first character of the first such substring is 
* returned. If it does not occur as a substring starting at <code>fromIndex</code> or beyond, 
* <code>-1</code> is returned.
*/
public final int getIndexOf(ByteString str) {
	return getIndexOf(str, 0);
}
/**
* Returns the index within this ByteString of the first occurrence of the specified substring,
* starting at the specified index.
* @param str a ByteString containing the substring to search for.
* @param fromIndex the index to start the search from.
* @return If the argument occurs as a substring within this object at a starting index no smaller 
* than <code>fromIndex</code>, then the index of the first character of the first such substring is 
* returned. If it does not occur as a substring starting at <code>fromIndex</code> or beyond, 
* <code>-1</code> is returned.
*/
public final int getIndexOf(ByteString str, int from) {
	return getIndexOf(str.bytValue, 0, str.intLength, from);
}
/**
* Returns the index within this ByteString of the first occurrence of the specified substring, starting at
* the specified index.
*
* @param comp a char array containing the substring to search for.
* @param off the first character in the array to compare
* @param len the total amount of characters to compare
* @param fromIndex the index to start the search from.
*
* @return If the argument occurs as a substring within this object at a starting index no smaller 
* than <code>fromIndex</code>, then the index of the first character of the first such substring is 
* returned. If it does not occur as a substring starting at <code>fromIndex</code> or beyond, 
* <code>-1</code> is returned.
*/
private final int getIndexOf(byte[] comp, int off, int len, int fromIndex) {

	// local copies
	byte[] charsThis = bytValue;
	int lenThis = intLength;

	// max value
	int lastIndexToCompare = lenThis - len;

	// check fromIndex
	if (fromIndex >= lenThis) {
		return -1;
	}
	if (fromIndex < 0) {
		fromIndex = 0;
	}
	if (len == 0) {
		return fromIndex;
	}

	// prepare loop
	byte first = comp[off];

	// loop for first char
	startSearchForFirstChar : while (true) {
		while (fromIndex <= lastIndexToCompare && charsThis[fromIndex] != first) {
			fromIndex++;
		}
		if (fromIndex > lastIndexToCompare) {
			return -1;
		}

		/* Found first character, now look at the rest of comp */
		int innerIndex = fromIndex + 1;
		int end = innerIndex + len - 1;
		int compIndex = off + 1;
		while (innerIndex < end) {
			if (charsThis[innerIndex++] != comp[compIndex++]) {
				fromIndex++;
				/* Look for str's first char again. */
				continue startSearchForFirstChar;
			}
		}
		return fromIndex;
		/* Found whole string. */
	}
}
/**
* Returns the index within this ByteString of the first occurrence of the specified substring.
* Cases are ignored.
*
* @param str a String containing the substring to search for.
*
* @return If the argument occurs as a substring within this object, then the index of the first character of
* the first such substring is returned. If it does not occur as a substring, <code>-1</code> is returned.
*/
public final int getIndexOfIgnoreCase(String str) {
	return getIndexOfIgnoreCase(str, 0);
}
/**
* Returns the index within this ByteString of the first occurrence of the specified substring,
* starting at the specified index.
*
* @param str a String containing the substring to search for.
* @param fromIndex the index to start the search from.
*
* @return If the argument occurs as a substring within this object at a starting index no smaller 
* than <code>fromIndex</code>, then the index of the first character of the first such substring is 
* returned. If it does not occur as a substring starting at <code>fromIndex</code> or beyond, 
* <code>-1</code> is returned.
*/
public final int getIndexOfIgnoreCase(String str, int fromIndex) {
	int len = str.length();
	return getIndexOfIgnoreCase(getASCII(str), 0, len, fromIndex);
}
/**
* Returns the index within this ByteString of the first occurrence of the specified substring,
* starting at the specified index.
*
* @param str a String containing the substring to search for.
* @param fromIndex the index to start the search from.
*
* @return If the argument occurs as a substring within this object at a starting index no smaller 
* than <code>fromIndex</code>, then the index of the first character of the first such substring is 
* returned. If it does not occur as a substring starting at <code>fromIndex</code> or beyond, 
* <code>-1</code> is returned.
*/
public final int getIndexOfIgnoreCase(ByteString str) {
	return getIndexOfIgnoreCase(str.bytValue, 0, str.intLength, 0);
}
/**
* Returns the index within this ByteString of the first occurrence of the specified substring,
* starting at the specified index.
*
* @param str a String containing the substring to search for.
* @param fromIndex the index to start the search from.
*
* @return If the argument occurs as a substring within this object at a starting index no smaller 
* than <code>fromIndex</code>, then the index of the first character of the first such substring is 
* returned. If it does not occur as a substring starting at <code>fromIndex</code> or beyond, 
* <code>-1</code> is returned.
*/
public final int getIndexOfIgnoreCase(ByteString str, int fromIndex) {
	return getIndexOfIgnoreCase(str.bytValue, 0, str.intLength, fromIndex);
}
/**
* Returns the index within this ByteString of the first occurrence of the specified substring,
* starting at the specified index.  Cases are ignored.
*
* @param comp a char array containing the substring to search for.
* @param off the first character in the array to compare
* @param len the total amount of characters to compare
* @param fromIndex the index to start the search from.
*
* @return If the argument occurs as a substring within this object at a starting index no smaller 
* than <code>fromIndex</code>, then the index of the first character of the first such substring is 
* returned. If it does not occur as a substring starting at <code>fromIndex</code> or beyond, 
* <code>-1</code> is returned.
*/
private final int getIndexOfIgnoreCase(byte[] comp, int off, int len, int fromIndex) {
	// local copies
	byte[] charsThis = bytValue;
	int lenThis = intLength;

	// max value
	int lastIndexToCompare = lenThis - len;

	// check fromIndex
	if (fromIndex >= lenThis) {
		return -1;
	}
	if (fromIndex < 0) {
		fromIndex = 0;
	}
	if (len == 0) {
		return fromIndex;
	}

	// prepare loop
	byte first = comp[off];
	byte firstUpper = (byte) (first & UPPER_BIT);
	byte firstLower = (byte) (first | LOWER_BIT);

	// loop for first char
	startSearchForFirstChar : while (true) {
		loop2 : while (fromIndex <= lastIndexToCompare) {
			byte loop;
			if (((loop = charsThis[fromIndex]) != first) && ((loop & UPPER_BIT) != firstUpper) && ((loop | LOWER_BIT) != firstLower))
				fromIndex++;
			else
				break loop2;
		}
		if (fromIndex > lastIndexToCompare) {
			return -1;
		}

		/* Found first character, now look at the rest of comp */
		int innerIndex = fromIndex + 1;
		int end = innerIndex + len - 1;
		int compIndex = off + 1;
		while (innerIndex < end) {
			byte loop;
			byte c = charsThis[innerIndex++];
			byte cUpper = (byte) (c & UPPER_BIT);
			byte cLower = (byte) (c | LOWER_BIT);
			if (((loop = comp[compIndex++]) != c) && (cUpper != (loop & UPPER_BIT)) && (cLower != (loop | LOWER_BIT))) {
				fromIndex++;
				/* Look for str's first char again. */
				continue startSearchForFirstChar;
			}
		}
		return fromIndex;
		/* Found whole string. */
	}
}
/**
* Returns the index within this ByteString of the last occurrence of the specified character.
* The ByteString is searched backwards starting at the last character.
* @param ch a character.
* @return the index of the last occurrence of the character in the character sequence represented by this
* object, or <code>-1</code> if the character does not occur.
*/
public final int getLastIndexOf(byte ch) {
	return getLastIndexOf(ch, intLength - 1);
}
/**
* Returns the index within this ByteString of the last occurrence of the specified character, searching 
* backward starting at the specified index.
* @param ch a character.
* @param fromIndex the index to start the search from.
* @return the index of the last occurrence of the character in the character sequence represented by this 
* object that is less than or equal to <code>fromIndex</code>, or <code>-1</code> if the character does not
* occur before that point.
*/
public final int getLastIndexOf(byte ch, int fromIndex) {
	byte[] v = bytValue;
	for (int i = ((fromIndex >= intLength) ? intLength - 1 : fromIndex); i >= 0; i--) {
		if (v[i] == ch) {
			return i;
		}
	}
	return -1;
}
/**
* Returns the index within this ByteString of the last occurrence of the specified substring.
*
* @param str the substring to search for.
*
* @return If the argument occurs one or more times as a substring within this object, then the index of the first character of the last such 
* substring is returned. If it does not occur as a substring, <code>-1</code> is returned.
*/
public final int getLastIndexOf(String str) {
	int len = str.length();
	return getLastIndexOf(str, intLength - 1);
}
/**
* Returns the index within this ByteString of the last occurrence of the specified substring.
* The returned index indicates the start of the substring, and it must be equal to or less than
* <code>fromIndex</code>.
*
* @param str the substring to search for.
* @param fromIndex the index to start the search from.
*
* @return If the argument occurs one or more times as a substring within this object at a starting
* index no greater than <code>fromIndex</code>, then the index of the first character of the last such 
* substring is returned. If it does not occur as a substring starting at <code>fromIndex</code> or earlier,
* <code>-1</code> is returned.
*/
public final int getLastIndexOf(String str, int fromIndex) {
	int len = str.length();
	return getLastIndexOf(getASCII(str), 0, len, fromIndex);
}
/**
* Returns the index within this ByteString of the last occurrence of the specified substring.
* The returned index indicates the start of the substring, and it must be equal to or less than
* <code>fromIndex</code>.
*
* @param str the substring to search for.
* @param fromIndex the index to start the search from.
*
* @return If the argument occurs one or more times as a substring within this object at a starting
* index no greater than <code>fromIndex</code>, then the index of the first character of the last such 
* substring is returned. If it does not occur as a substring starting at <code>fromIndex</code> or earlier,
* <code>-1</code> is returned.
*/
public final int getLastIndexOf(ByteString str) {
	return getLastIndexOf(str, 0);
}
/**
* Returns the index within this ByteString of the last occurrence of the specified substring.
* The returned index indicates the start of the substring, and it must be equal to or less than
* <code>fromIndex</code>.
*
* @param str the substring to search for.
* @param fromIndex the index to start the search from.
*
* @return If the argument occurs one or more times as a substring within this object at a starting
* index no greater than <code>fromIndex</code>, then the index of the first character of the last such 
* substring is returned. If it does not occur as a substring starting at <code>fromIndex</code> or earlier,
* <code>-1</code> is returned.
*/
public final int getLastIndexOf(ByteString str, int fromIndex) {
	return getLastIndexOf(str.bytValue, 0, str.intLength, fromIndex);
}
/**
* Returns the index within this ByteString of the last occurrence of the specified substring.
* The returned index indicates the start of the substring, and it must be equal to or less than
* <code>fromIndex</code>.
*
* @param chars the substring to search for.
* @param off the first char to compare
* @param len the total number of chars in 'chars' to compare.
* @param fromIndex the index to start the search from.
*
* @return If the argument occurs one or more times as a substring within this object at a starting
* index no greater than <code>fromIndex</code>, then the index of the first character of the last such 
* substring is returned. If it does not occur as a substring starting at <code>fromIndex</code> or earlier,
* <code>-1</code> is returned.
*/
private final int getLastIndexOf(byte[] chars, int off, int len, int fromIndex) {
	/* 
	 * Check arguments; return immediately where possible. For
	 * consistency, don't check for null str.
	 */
	if (fromIndex < 0) {
		return -1;
	}
	{
		int t = intLength - len;
		if (fromIndex > t) {
			fromIndex = t;
		}
	}
	/* Empty string always matches. */
	if (len == 0) {
		return fromIndex;
	}

	//
	byte[] val = bytValue;
	int strLastIndex = off + len - 1;
	byte strLastChar = chars[strLastIndex];
	int lastCharPos = len - 1;
	int lastCharIndex = lastCharPos + fromIndex;

	// loop
	startSearchForLastChar : while (true) {

		/* Look for the last character */
		while (lastCharIndex >= lastCharPos && val[lastCharIndex] != strLastChar) {
			lastCharIndex--;
		}
		if (lastCharIndex < lastCharPos) {
			return -1;
		}

		/* Found last character, now look at the rest of chars. */
		int thisIndex = lastCharIndex - 1;
		int start = thisIndex - (len - 1);
		int paramIndex = strLastIndex - 1;
		while (thisIndex > start) {
			if (val[thisIndex--] != chars[paramIndex--]) {
				lastCharIndex--;
				/* Look for str's last char again. */
				continue startSearchForLastChar;
			}
		}
		return start + 1;
		/* Found whole string. */
	}
}
/**
* Returns the index within this ByteString of the last occurrence of the specified substring.
*
* @param str the substring to search for.
*
* @return If the argument occurs one or more times as a substring within this object, then the index of the first character of the last such 
* substring is returned. If it does not occur as a substring, <code>-1</code> is returned.
*/
public final int getLastIndexOfIgnoreCase(String str) {
	int len = str.length();
	return getLastIndexOfIgnoreCase(str, intLength - 1);
}
/**
* Returns the index within this ByteString of the last occurrence of the specified substring.
* The returned index indicates the start of the substring, and it must be equal to or less than
* <code>fromIndex</code>.
*
* @param str the substring to search for.
* @param fromIndex the index to start the search from.
*
* @return If the argument occurs one or more times as a substring within this object at a starting
* index no greater than <code>fromIndex</code>, then the index of the first character of the last such 
* substring is returned. If it does not occur as a substring starting at <code>fromIndex</code> or earlier,
* <code>-1</code> is returned.
*/
public final int getLastIndexOfIgnoreCase(String str, int fromIndex) {
	int len = str.length();
	return getLastIndexOfIgnoreCase(getASCII(str), 0, len, fromIndex);
}
/**
* Returns the index within this ByteString of the last occurrence of the specified substring.
* The returned index indicates the start of the substring, and it must be equal to or less than
* <code>fromIndex</code>.
*
* @param str the substring to search for.
* @param fromIndex the index to start the search from.
*
* @return If the argument occurs one or more times as a substring within this object at a starting
* index no greater than <code>fromIndex</code>, then the index of the first character of the last such 
* substring is returned. If it does not occur as a substring starting at <code>fromIndex</code> or earlier,
* <code>-1</code> is returned.
*/
public final int getLastIndexOfIgnoreCase(ByteString str) {
	return getLastIndexOfIgnoreCase(str, 0);
}
/**
* Returns the index within this ByteString of the last occurrence of the specified substring.
* The returned index indicates the start of the substring, and it must be equal to or less than
* <code>fromIndex</code>.
*
* @param str the substring to search for.
* @param fromIndex the index to start the search from.
*
* @return If the argument occurs one or more times as a substring within this object at a starting
* index no greater than <code>fromIndex</code>, then the index of the first character of the last such 
* substring is returned. If it does not occur as a substring starting at <code>fromIndex</code> or earlier,
* <code>-1</code> is returned.
*/
public final int getLastIndexOfIgnoreCase(ByteString str, int fromIndex) {
	return getLastIndexOfIgnoreCase(str.bytValue, 0, str.intLength, fromIndex);
}
/**
* Returns the index within this ByteString of the last occurrence of the specified substring.
* The returned index indicates the start of the substring, and it must be equal to or less than
* <code>fromIndex</code>.
*
* @param chars the substring to search for.
* @param off the first char to compare
* @param len the total number of chars in 'chars' to compare.
* @param fromIndex the index to start the search from.
* @param ignoreCase if true, the case is ignored
*
* @return If the argument occurs one or more times as a substring within this object at a starting
* index no greater than <code>fromIndex</code>, then the index of the first character of the last such 
* substring is returned. If it does not occur as a substring starting at <code>fromIndex</code> or earlier,
* <code>-1</code> is returned.
*/
private final int getLastIndexOfIgnoreCase(byte[] chars, int off, int len, int fromIndex) {
	/* 
	 * Check arguments; return immediately where possible. For
	 * consistency, don't check for null str.
	 */
	if (fromIndex < 0) {
		return -1;
	}
	{
		int t = intLength - len;
		if (fromIndex > t) {
			fromIndex = t;
		}
	}
	/* Empty string always matches. */
	if (len == 0) {
		return fromIndex;
	}

	//
	byte[] val = bytValue;
	int lastIndex = off + len - 1;
	byte last = chars[lastIndex];
	byte lastUpper = (byte) (last & UPPER_BIT);
	byte lastLower = (byte) (last | LOWER_BIT);
	int lastSearchPos = len - 1;
	int lastSearchIndex = lastSearchPos + fromIndex;

	// loop
	startSearchForLastChar : while (true) {

		/* Look for the last character */
		loop2 : while (lastSearchIndex >= lastSearchPos) {
			byte loop;
			if (((loop = val[lastSearchIndex]) != last) && ((loop & UPPER_BIT) != lastUpper) && ((loop | LOWER_BIT) != lastLower))
				lastSearchIndex--;
			else
				break loop2;
		}
		if (lastSearchIndex < lastSearchPos) {
			return -1;
		}

		/* Found last character, now look at the rest of chars. */
		int thisIndex = lastSearchIndex - 1;
		int start = thisIndex - (len - 1);
		int paramIndex = lastIndex - 1;
		while (thisIndex > start) {
			byte loop;
			byte c = val[thisIndex--];
			byte cUpper = (byte) (c & UPPER_BIT);
			byte cLower = (byte) (c | LOWER_BIT);
			if (((loop = chars[paramIndex--]) != c) && ((loop & UPPER_BIT) != cUpper) && ((loop | LOWER_BIT) != cLower)) {
				lastSearchIndex--;
				/* Look for str's last char again. */
				continue startSearchForLastChar;
			}
		}
		return start + 1;
		/* Found whole string. */
	}
}
/**
* Returns the length (character count) of this ByteString.
* @return returns the number of characters in this ByteString.
*/
public final int getLength() {
	return intLength;
}
/**
* sets a line break specific for this object.  Since the primary goal of the class is the handling of
* text from internet protocols, the default is set to CRLF.
* @param lb the line seperator string
*/
public String getLineBreak() {
	return new String(bytLineBreak);
}
/**
* returns the number of lines in this ByteString by counting the number of hard breaks.
* @return returns the number of lines, 0 if the buffer is empty
*/
public final int getLineCount() {
	if (intLength == 0)
		return 0;
	byte lb = bytLineBreak[0];
	int count = getCharCount(lb, 0, intLength - 1);
	// if the last character is not a line feed, there is one more line.
	int lbLen = bytLineBreak.length;
	if ((bytValue[intLength - lbLen]) != lb)
		count++;
	return count;
}
/**
* returns a String containing all chars between the given position <code>fromIndex</code>, inclusive, and
* the next line separator, exclusive.
* After a call to this method the property <code>nextLineStart</code>&nbsp; points to the first character of
* the next line.
* @return returns a ByteString containing the line of text.
* @param offset the first character to parse into the line.
* @param skipSpaces when true, all leading and trailing white spaces (characters with unicode values
* lower than 0x20) are removed from the resulting ByteString.  Any row of consecutive white spaces are
* replaced by a single space character.
*/
public final ByteString getLineHardBreak(int offset) {
	if (offset < 0 || offset > intLength)
		throw new StringIndexOutOfBoundsException("offset: " + offset);

	// avoid field lookup
	byte[] source = bytValue;
	int len = intLength;

	// find line end.  since we do not care for the length of a line, we simply look up the
	// line break
	int lineEnd = getIndexOf(bytLineBreak[0], offset);

	// if there is none, we return the whole string to the end
	if (lineEnd == -1) {
		lineEnd = len;
	}

	// calc line length
	int lineLength = lineEnd - offset;

	// don't need much calculation, if the white spaces are not skipped
	intNextLineStart = offset + lineLength + bytLineBreak.length;
	return new ByteString(source, offset, lineLength);
}
/**
* returns an Enumeration which returns single lines.  A line is either the complete text or every row
* of consecutive chars that is terminated by a line break.  The line break is not included in the resulting
* ByteString.
* @return returns an Enumeration
*/
public final Enumeration getLineHardBreakEnumeration() {
	return new LineEnumeration();
}
/**
* returns the character position where the next line will start.  This value has a meaningful value only
* when parsing lines with the getLineXXX methods without changing the content of the buffer at the same
* time
* @return returns the character position where the next line will start.
* @see setNextLineStart
* @see getLineFixedLength
*/
public final int getNextLineStart() {
	return intNextLineStart;
}
/**
* returns a String representing the bytes of the byte array argument as low byte.  The high byte is 0.
* @param b a byte array
* @return returns a String representing the bytes fo the argument array.
*/
public final static String getString(byte[] b) {
	return getString(b, 0, b.length);
}
/**
* returns a String containing the bytes of the byte array argument as low byte.  The high byte is 0.
* @param b a byte array
* @param off the index of the first byte in the argument array to be converted.
* @param len the total number of bytes to be converted.
* @return returns String representing the bytes of the argument
*/
public final static String getString(byte[] b, int off, int len) {
	char[] result = new char[len];
	for (int i = len; i-- > 0;) {
		result[i] = (char) (b[i + off] & 0xff);
	}
	return new String(result);
}
/**
* Returns a new string that is a substring of this ByteString. The substring begins at the specified
* index and extends to the end of this ByteString.
* @param begin the beginning index, inclusive.
* @return the specified substring.
*/
public final ByteString getSubstring(int begin) {
	return getSubstring(begin, intLength);
}
/**
* Returns a new string that is a substring of this ByteString. The substring begins at the specified 
* <code>begin</code> index and extends to the character at index <code>end - 1</code>. 
* @param begin the beginning index, inclusive.
* @param end the ending index, exclusive.
* @return the specified substring.
*/
public final ByteString getSubstring(int begin, int end) {
	if (begin < 0 || end < 0)
		throw new StringIndexOutOfBoundsException(begin + "," + end);
	if (begin > end) {
		throw new StringIndexOutOfBoundsException(end - begin);
	}
	if (end > intLength)
		end = intLength;
	if (begin > intLength)
		begin = intLength;
	return new ByteString(bytValue, begin, end - begin);
}
/**
* Returns a hashcode for this ByteString.
* @return a hash code value for this object. 
*/
public final int hashCode() {
	int h = 0;
	int off = 0;
	byte[] val = bytValue;
	int len = intLength;
	if (len < 16) {
		for (int i = len; i > 0; i--) {
			h = (h * 37) + val[off++];
		}
	} else {
		// only sample some characters
		int skip = len / 8;
		for (int i = len; i > 0; i -= skip, off += skip) {
			h = (h * 39) + val[off];
		}
	}
	return h;
}
/**
* can reactivate a ByteString after a dispose()
* @param size the new size.  If the object has not disposed, is it ensured, that the buffer can hold
* at least this number of chars
*/
public void init(int size) {
	if (bytValue == null) {
		bytValue = new byte[size + 16];
	} else {
		ensureCapacity(size);
	}
}
/**
* Inserts the <code>char</code> argument into this ByteString. 
* <p>The second argument is inserted into the contents of this ByteString at the position indicated by
* <code>offset</code>. The length of this string buffer increases by one. 
* <p>The offset argument must be greater than or equal to <code>0</code>, and less than or equal to the
* length of this string buffer. 
* @param offset the offset.
* @param ch a <code>char</code>.
* @return returns this ByteString.
*/
public final ByteString insert(int offset, byte c) {
	int newcount = intLength + 1;
	if (newcount > bytValue.length)
		ensureCapacity(newcount);
	System.arraycopy(bytValue, offset, bytValue, offset + 1, intLength - offset);
	bytValue[offset] = c;
	intLength = newcount;
	return this;
}
/**
* Inserts the <code>char</code> argument into this ByteString. 
* <p>The second argument is inserted into the contents of this ByteString at the position indicated by
* <code>offset</code>. The length of this string buffer increases by one. 
* <p>The offset argument must be greater than or equal to <code>0</code>, and less than or equal to the
* length of this string buffer. 
* @param offset the offset.
* @param ch a <code>char</code>.
* @return returns this ByteString.
*/
public final ByteString insert(int offset, char c) {
	int newcount = intLength + 1;
	if (newcount > bytValue.length)
		ensureCapacity(newcount);
	System.arraycopy(bytValue, offset, bytValue, offset + 1, intLength - offset);
	bytValue[offset] = (byte) c;
	intLength = newcount;
	return this;
}
/**
* Inserts the string representation of the <code>double</code> argument into this ByteString. 
* <p>The second argument is converted to a string as if by the method <code>String.valueOf</code>, and the
* characters of that string are then inserted into this ByteString at the indicated offset. 
* <p>The offset argument must be greater than or equal to <code>0</code>, and less than or equal to the
* length of this ByteString. 
*
* @param offset the offset.
* @param b a <code>double</code>.
* @return this ByteString.
*/
public final ByteString insert(int offset, double d) {
	return insert(offset, String.valueOf(d));
}
/**
* Inserts the string representation of the <code>float</code> argument into this ByteString. 
* <p>The second argument is converted to a string as if by the method <code>String.valueOf</code>, and the
* characters of that string are then inserted into this ByteString at the indicated offset. 
* <p>The offset argument must be greater than or equal to <code>0</code>, and less than or equal to the
* length of this ByteString. 
*
* @param offset the offset.
* @param b a <code>float</code>.
* @return this ByteString.
*/
public final ByteString insert(int offset, float f) {
	return insert(offset, String.valueOf(f));
}
/**
* Inserts the string representation of the second <code>int</code> argument into this ByteString. 
* <p>The second argument is converted to a string as if by the method <code>String.valueOf</code>, and the
* characters of that string are then inserted into this ByteString at the indicated offset. 
* <p>The offset argument must be greater than or equal to <code>0</code>, and less than or equal to the
* length of this ByteString. 
*
* @param offset the offset.
* @param b an <code>int</code>.
* @return returns this ByteString.
*/
public final ByteString insert(int offset, int i) {
	// The calculation is done in the positive and Integer.MIN_VALUE has no positive expression
	if (i == Integer.MIN_VALUE) {
		return insert(offset, INTEGER_MIN, 0, INTEGER_MIN.length);
	}
	byte[] buf = new byte[12];
	boolean negative = (i < 0);
	int charPos = 12;
	if (negative) {
		i = -i;
	}
	if (i < 10 && !negative) {
		return insert(offset, (byte) (i + 48));
	}
	do {
		int digit = i % 100;
		buf[--charPos] = UNITS[digit];
		buf[--charPos] = TENTH_UNITS[digit];
		i = i / 100;
	} while (i != 0);
	// if the last 100 block had only one digit, go one back
	if (buf[charPos] == 48) {
		charPos++;
	}
	if (negative) {
		buf[--charPos] = MINUS;
	}
	return insert(offset, buf, charPos, (12 - charPos));
}
/**
* Inserts the string representation of the <code>long</code> argument into this ByteString.
* <p>The second argument is converted to a string as if by the method <code>String.valueOf</code>, and the
* characters of that string are then inserted into this ByteString at the indicated offset. 
* <p>The offset argument must be greater than or equal to <code>0</code>, and less than or equal to the
* length of this ByteString. 
*
* @param offset the offset.
* @param b a <code>long</code>.
* @return this ByteString.
*/
public final ByteString insert(int offset, long l) {
	// The calculation is done in the positive and Long.MIN_VALUE has no positive expression
	if (l == Long.MIN_VALUE) {
		return insert(offset, LONG_MIN, 0, LONG_MIN.length);
	}
	byte[] buf = new byte[20];
	boolean negative = (l < 0);
	int charPos = 20;
	if (negative) {
		l = -l;
	}
	if (l < 10 && !negative) {
		return insert(offset, (byte) (l + 48));
	}
	do {
		int digit = (int) (l % 100);
		buf[--charPos] = UNITS[digit];
		buf[--charPos] = TENTH_UNITS[digit];
		l = l / 100;
	} while (l != 0);
	// if the last 100 block had only one digit, go one back
	if (buf[charPos] == 48) {
		charPos++;
	}
	if (negative) {
		buf[--charPos] = MINUS;
	}
	return insert(offset, buf, charPos, (20 - charPos));
}
/**
* Inserts the string representation of the <code>Object</code> argument into this ByteString. 
* <p>The second argument is converted to a string as if by the method <code>String.valueOf</code>, and the
* characters of that string are then inserted into this ByteString at the indicated offset. 
* <p>The offset argument must be greater than or equal to <code>0</code>, and less than or equal to the 
* length of this ByteString. 
*
* @param offset the offset.
* @param b an <code>Object</code>.
* @return this ByteString.
*/
public final ByteString insert(int offset, Object obj) {
	return insert(offset, String.valueOf(obj));
}
/**
* Inserts the string into this ByteString. 
* <p>The characters of the <code>String</code> argument are inserted, in order, into this ByteString at
* the indicated offset. The length of this ByteString is increased by the length of the argument.
* <p>The offset argument must be greater than or equal to <code>0</code>, and less than or equal to the
* length of this ByteString.
* @param offset the offset.
* @param str a string.
* @return this ByteString.
*/
public final ByteString insert(int offset, String str) {
	int len = str.length();
	int newcount = intLength + len;
	if (newcount > bytValue.length)
		ensureCapacity(newcount);
	System.arraycopy(bytValue, offset, bytValue, offset + len, intLength - offset);
	System.arraycopy(getASCII(str), 0, bytValue, offset, len);
	intLength = newcount;
	return this;
}
/**
* Inserts the string representation of the <code>boolean</code> argument into this ByteString. 
* <P>The offset argument must be greater than or equal to <code>0</code>, and less than or equal to the 
* length of this ByteString. 
* @param offset the offset.
* @param b a <code>boolean</code>.
* @return returns this ByteString.
*/
public final ByteString insert(int offset, boolean b) {
	return insert(offset, String.valueOf(b));
}
/**
* Inserts the string representation of the <code>char</code> array argument into this ByteString. 
* <p>The characters of the array argument are inserted into the contents of this ByteString at the
* position indicated by <code>offset</code>. The length of this ByteString increases by the length of 
* the argument. 
* @param offset the offset.
* @param str a character array.
* @param off the first character in the argument array to insert
* @param len the total amount of characters to insert.
* @return this ByteString.
*/
public final ByteString insert(int start, byte[] str, int off, int len) {
	int newcount = intLength + len;
	if (newcount > bytValue.length)
		ensureCapacity(newcount);
	System.arraycopy(bytValue, start, bytValue, start + len, intLength - start);
	System.arraycopy(str, off, bytValue, start, len);
	intLength = newcount;
	return this;
}
/**
* Inserts the string representation of the <code>char</code> array argument into this ByteString. 
* <p>The characters of the array argument are inserted into the contents of this ByteString at the
* position indicated by <code>offset</code>. The length of this ByteString increases by the length of 
* the argument. 
* @param offset the offset.
* @param ch a character array.
* @return this ByteString.
*/
public final ByteString insert(int offset, char[] str) {
	return insert(offset, str, 0, str.length);
}
/**
* Inserts the string representation of the <code>char</code> array argument into this ByteString. 
* <p>The characters of the array argument are inserted into the contents of this ByteString at the
* position indicated by <code>offset</code>. The length of this ByteString increases by the length of 
* the argument. 
* @param offset the offset.
* @param str a character array.
* @param off the first character in the argument array to insert
* @param len the total amount of characters to insert.
* @return this ByteString.
*/
public final ByteString insert(int start, char[] str, int off, int len) {
	return insert(start, getASCII(str), off, len);
}
/**
* returns true, if this ByteString conatins no chars
* @return returns true, if this ByteString conatins no chars
*/
public final boolean isEmpty() {
	return intLength == 0;
}
/**
* Compares the content of a char array lexicographically to this ByteString. The comparison is
* based on the byte value of each character.  This method compares the content of the argument to a
* region in this ByteString of the same length, starting at a given position.  It returns true,
* if there are at least len characters in this ByteString, counting from the start position to
* the end, and all characters are equal to those in the argument.
* @param str String to be compared.
* @param start the first character of this ByteString to compare
* @param ignoreCase if true, the case is ignored.
* @return returns true, if the two regions match, false otherwise.
*/
public final boolean isMatchingRegion(String str, int start, boolean ignoreCase) {
	return isMatchingRegion(getASCII(str), 0, str.length(), start, ignoreCase);
}
/**
* Compares the content of a char array lexicographically to this ByteString. The comparison is
* based on the byte value of each character.  This method compares the content of the argument to a
* region in this ByteString of the same length, starting at a given position.  It returns true,
* if there are at least len characters in this ByteString, counting from the start position to
* the end, and all characters are equal to those in the argument.
* @param str String to be compared.
* @param start the first character of this ByteString to compare
* @param ignoreCase if true, the case is ignored.
* @return returns true, if the two regions match, false otherwise.
*/
public final boolean isMatchingRegion(ByteString str, int start, boolean ignoreCase) {
	return isMatchingRegion(str.bytValue, 0, str.intLength, start, ignoreCase);
}
/**
* Compares the content of a char array lexicographically to this ByteString. The comparison is based on the
* byte value of each character.  This method compares the content of the argument to a region in this
* ByteString of the same length, starting at a given position.  It returns true, if there are at least len
* characters in this ByteString, counting from the start position to the end, and all characters are equal
* to those in the argument.
*
* @param comp char array to be compared.
* @param off the first character of the argument to compare
* @param len the total number of chars to compare
* @param start the first character of this ByteString to compare
* @param ignoreCase if true, the case is ignored.
*
* @return returns true, if the two regions match, false otherwise.
*/
private final boolean isMatchingRegion(byte[] chars, int off, int len, int start, boolean ignoreCase) {
	if (off < 0 || off > chars.length || len < 0 || len + off > chars.length || start > intLength)
		throw new IllegalArgumentException();
	if (intLength - start < len)
		return false;

	//
	byte[] thisChars = bytValue;
	int indexThis = start;
	int indexThat = off;
	while (len-- != 0) {
		byte cthis = thisChars[indexThis++];
		byte cthat = chars[indexThat++];
		if (cthis == cthat)
			continue;
		if (ignoreCase && (cthis & UPPER_BIT) == (cthat & UPPER_BIT)) {
			continue;
		}
		return false;
	}
	return true;
}
/**
* overwrites characters from the given start position with the argument char.
* @param ch the char to put into the buffer
* @param start the start position in the ByteString.
* @param end the last character to overwrite
* @return returns this ByteString
*/
public final ByteString overwrite(byte ch, int start, int end) {
	if (start < 0)
		start = 0;
	if (end < 0)
		return this;
	if (start >= intLength)
		return this;
	if (end >= intLength)
		end = intLength - 1;
	byte[] chars = bytValue;
	for (int i = start; i <= end; i++) {
		chars[i] = ch;
	}
	return this;
}
/**
* overwrites characters from the given start position with the argument char.
* @param ch the char to put into the ByteString
* @param start the start position in the buffer.
* @param end the last character to overwrite
* @return returns this ByteString
*/
public final ByteString overwrite(char ch, int start, int end) {
	return overwrite((byte) ch, start, end);
}
/**
* replaces all redundant white spaces in this this ByteString with on single space
* @return returns this ByteString
*/
public final ByteString reduceSpaces() {
	int oldLen = intLength;
	int oldPointer = 0;
	int newPointer = 0;
	byte[] values = bytValue;
	boolean inWord = false;
	while (oldPointer < oldLen) {
		int ch = values[oldPointer++] & 0x000000ff;
		if (ch > 32) {
			values[newPointer++] = (byte) ch;
			inWord = true;
		} else {
			if (inWord) {
				values[newPointer++] = 32;
				inWord = false;
			}
		}
	}
	intLength = newPointer;
	return this;
}
/**
* replaces all redundant white spaces in this this ByteString with on single space, but keeps
* line breaks.
* @param keepLineBreaks if true, line breaks are not replaced by spaces
* @return returns this ByteString
*/
public final ByteString reduceSpaces(boolean keepLineBreaks) {
	int oldLen = intLength;
	int oldPointer = 0;
	int newPointer = 0;
	byte[] values = bytValue;
	boolean inWord = false;
	while (oldPointer < oldLen) {
		int ch = values[oldPointer++] & 0x000000ff;
		if (ch > 32 || (keepLineBreaks && (ch == 13 || ch == 10))) {
			values[newPointer++] = (byte) ch;
			if (ch != 13 && ch != 10) {
				inWord = true;
			}
		} else {
			if (inWord) {
				values[newPointer++] = 32;
				inWord = false;
			}
		}
	}
	intLength = newPointer;
	return this;
}
/**
* Replaces all occurrences of <code>oldChar</code> in this ByteString with <code>newChar</code>. 
* @param oldChar the old character.
* @param newChar the new character.
* @return returns this ByteString
*/
public final ByteString replace(byte oldChar, byte newChar) {
	if (oldChar != newChar) {
		int len = intLength;
		byte[] val = bytValue;
		for (int i = len; i-- > 0;) {
			if (val[i] == oldChar) {
				val[i] = newChar;
			}
		}
	}
	return this;
}
/**
* replaces all occurances of the first String with the second.
* @param oldStr the old string to replace
* @param newStr the String to replace old with
* @param ignoreCase ignore case when looking for the first string in this ByteString, however, the
* second char is inserted as is.
* @return returns this ByteString
*/
public final ByteString replace(String oldStr, String newStr, boolean ignoreCase) {
	int oldLen = oldStr.length();
	int newLen = newStr.length();
	return replace(getASCII(oldStr), 0, oldLen, getASCII(newStr), 0, newLen, ignoreCase);
}
/**
* replaces all occurances of the first String with the second.
* @param oldStr the old string to replace
* @param newStr the String to replace old with
* @param ignoreCase ignore case when looking for the first string in this ByteString, however, the
* second char is inserted as is.
* @return returns this ByteString
*/
public final ByteString replace(ByteString oldStr, ByteString newStr, boolean ignoreCase) {
	return replace(oldStr.bytValue, 0, oldStr.intLength, newStr.bytValue, 0, newStr.intLength, ignoreCase);
}
/**
* replaces all occurances of the first char array with the second.
* @param oldStr the old chars to replace
* @param oldOff the first char of the string to replace
* @param oldLen the total number of chars to replace, must be > 0
* @param newStr the new chars to replace the old with
* @param newOff the first char in the new string
* @param newLen the total number of chars to, must be > 0
* @param ignoreCase ignore case when looking for the first char in this ByteString, however, the
* second char is inserted as is.
* @return returns this ByteString
*/
private final ByteString replace(byte[] oldStr, int oldOff, int oldLen, byte[] newStr, int newOff, int newLen, boolean ignoreCase) {
	if (oldStr == null || newStr == null)
		throw new NullPointerException();
	if (oldOff < 0 || oldLen < 0 || newOff < 0 || newLen < 0 || oldOff + oldLen > oldStr.length || newOff + newLen > newStr.length)
		throw new StringIndexOutOfBoundsException();

	// nothing to do?
	if (oldLen == 0)
		return this;

	// go for it
	int pos = 0;
	while ((pos = (ignoreCase ? getIndexOfIgnoreCase(oldStr, oldOff, oldLen, pos) : getIndexOf(oldStr, oldOff, oldLen, pos))) != -1) {
		if (oldLen == newLen) {
			System.arraycopy(newStr, newOff, bytValue, pos, newLen);
		} else {
			if (oldLen > newLen) {
				if (newLen > 0) {
					System.arraycopy(newStr, newOff, bytValue, pos, newLen);
				}
				delete(pos + newLen, oldLen - newLen);
			} else {
				System.arraycopy(newStr, newOff, bytValue, pos, oldLen);
				insert(pos + oldLen, newStr, newOff + oldLen, newLen - oldLen);
			}
		}
		pos += newLen;
	}
	return this;
}
/**
* The character sequence contained in this ByteString is replaced by the reverse of the sequence. 
* @return returns this ByteString.
*/
public final ByteString reverse() {
	int lastChar = intLength - 1;
	for (int i = (lastChar - 1) >> 1; i >= 0; --i) {
		byte temp = bytValue[i];
		bytValue[i] = bytValue[lastChar - i];
		bytValue[lastChar - i] = temp;
	}
	return this;
}
/**
* sets the 'real' char array in this ByteString.
* @param chars a char array
* @param len the number of chars, beginning at position <code>0</code>.  This value must not be greater 
* than the length of the array.
*/
public final void setByteArray(byte[] chars, int len) {
	if (chars == null)
		throw new NullPointerException();
	if (len > chars.length || len < 0)
		throw new StringIndexOutOfBoundsException(len);
	bytValue = chars;
	intLength = len;
}
/**
* The character at the specified index of this ByteString is set to <code>ch</code>. 
* <p>The offset argument must be greater than or equal to <code>0</code>, and less than the length of this
* ByteString.
* @param index the index of the character to modify.
* @param ch the new character.
*/
public final void setCharAt(int index, byte ch) {
	if ((index < 0) || (index >= intLength))
		throw new StringIndexOutOfBoundsException(index);
	bytValue[index] = ch;
}
/**
* The character at the specified index of this ByteString is set to <code>ch</code>. 
* <p>The offset argument must be greater than or equal to <code>0</code>, and less than the length of this
* ByteString. 
* @param index the index of the character to modify.
* @param ch the new character.
*/
public final void setCharAt(int index, char ch) {
	setCharAt(index, (byte) ch);
}
/**
* Sets the length of this ByteString.
* If the <code>newLength</code> argument is less than the current length of the ByteString, the ByteString
* is truncated to contain exactly the number of characters given by the <code>newLength</code> argument.
* <p>If the <code>newLength</code> argument is greater than or equal to the current length, sufficient null
* characters (<code>'&#92;u0000'</code>) are appended to the ByteString so that length becomes the
* <code>newLength</code> argument.
* <p>The <code>newLength</code> argument must be greater than or equal to <code>0</code>. 
* @param newLength the new length of the ByteString.
*/
public final void setLength(int newLength) {
	if (newLength < 0)
		throw new StringIndexOutOfBoundsException(newLength);
	if (newLength > bytValue.length)
		ensureCapacity(newLength);
	if (intLength < newLength) {
		for (; intLength < newLength; intLength++) {
			bytValue[intLength] = 0;
		}
	} else {
		intLength = newLength;
	}
}
/**
* Sets the length of this ByteString.  This method does not change the internal byte array.  It should only
* be used after the content of the char array has been changed 'from the outside'.
* @param newLength the new length of the buffer.  Must not be greater than <code>getCapacity()</code>.
*/
public final void setLengthIndex(int newLength) {
	if (newLength < 0 || newLength > bytValue.length)
		throw new StringIndexOutOfBoundsException(newLength);
	intLength = newLength;
}
/**
* sets a line break specific for this object.  Since the primary goal of the class is the handling of
* text from internet protocols, the default is set to CRLF.
* @param lb the line seperator string
*/
public void setLineBreak(String lb) {
	bytLineBreak = lb.getBytes();
}
/**
* sets the character position where the next line will start.  Since the string buffer only sets this value
* after calculating a line with any getLineXXX() method, this method will not effect any future line
* parsing.  However, it is useful when manipulating the buffer from different methods where new lines are
* read by first reading this value.
* @param the character position where the next line starts.
* @see getNextLineStart
* @see getLineFixedLength
* @see getLineSoftLength
* @see getLineHardBreak
*/
public final void setNextLineStart(int lineNo) {
	intNextLineStart = lineNo;
}
/**
* Tests if this ByteString starts with the specified prefix.
* @param str the prefix.
* @return <code>true</code> if the character sequence represented by the argument is a praefix of the
* character sequence represented by this object; <code>false</code> otherwise.
*/
public final boolean startsWith(String str) {
	int len = str.length();
	if (intLength < len)
		return false;
	return isMatchingRegion(getASCII(str), 0, len, 0, false);
}
/**
* removes all white spaces from this ByteString
* @return returns this ByteString
*/
public final ByteString stripSpaces() {
	int oldLen = intLength;
	int oldPointer = 0;
	int newPointer = 0;
	byte[] values = bytValue;
	while (oldPointer < oldLen) {
		byte ch = values[oldPointer++];
		if ((ch & 0x000000ff) > 32) {
			values[newPointer++] = ch;
		}
	}
	intLength = newPointer;
	return this;
}
/**
* Converts this ByteString to a new byte array.
* @param off the first char to copy into the new byte array
* @param the total amount of characters to copy.
* @return a newly allocated byte array whose length is <code>len</code> and whose contents are initialized
* to contain the character sequence represented by this ByteString, beginning at <code>off</code>.
*/
public final byte[] toByteArray() {
	return toByteArray(0, intLength);
}
/**
* Converts this ByteString to a new byte array.
* @param off the first char to copy into the new byte array
* @param the total amount of characters to copy.
* @return a newly allocated byte array whose length is <code>len</code> and whose contents are initialized
* to contain the character sequence represented by this ByteString, beginning at <code>off</code>.
*/
public final byte[] toByteArray(int off, int len) {
	if (off < 0 || len < 0)
		throw new StringIndexOutOfBoundsException(off + "," + len);
	if (off >= intLength)
		off = intLength;
	if (off + len > intLength)
		len = intLength - off;
	byte result[] = new byte[len];
	System.arraycopy(bytValue, off, result, 0, len);
	return result;
}
/**
* Converts this ByteString to a new character array.
* @return a newly allocated character array whose length is the length of this ByteString and whose
* contents are initialized to contain the character sequence represented by this ByteString.
*/
public final char[] toCharArray() {
	return toCharArray(0, intLength);
}
/**
* Converts this ByteString to a new character array.
* @param off the first char to copy into the new char array
* @param the total amount of characters to copy.
* @return a newly allocated character array whose length is <code>len</code> and whose contents are 
* initialized to contain the character sequence represented by this string buffer, beginning at <code>off</code>.
*/
public final char[] toCharArray(int off, int len) {
	if (off < 0 || len < 0)
		throw new StringIndexOutOfBoundsException(off + "," + len);
	if (off >= intLength)
		off = intLength;
	if (off + len > intLength)
		len = intLength - off;
	return getCharArray(bytValue, off, len);
}
/**
* Converts to a string representing the data in this ByteString.
* @param off the first char to be copied into the new String
* @param len the total amount of characters to be copied.
* @return a string representation of the ByteString.
*/
public final String toString() {
	return getString(bytValue, 0, intLength);
}
/**
* Removes white space from both ends of this ByteString. 
* <p>All characters that have codes less than or equal to <code>'&#92;u0020'</code> (the space character)
* are considered to be white space. 
* @return this ByteString
*/
public final ByteString trim() {
	int len = intLength;
	int start = 0;
	byte[] val = bytValue;
	while ((start < len) && ((val[start] & 0x000000ff) <= ' ')) {
		start++;
	}
	while ((start < len) && ((val[len - 1] & 0x000000ff) <= ' ')) {
		len--;
	}
	System.arraycopy(val, start, val, 0, intLength = len - start);
	return this;
}
/**
* Removes white space from the beginning of this ByteString. 
* <p>All characters that have codes less than or equal to <code>'&#92;u0020'</code> (the space character)
* are considered to be white space. 
* @return this ByteString
*/
public final ByteString trimLeft() {
	int len = intLength;
	int start = 0;
	byte[] val = bytValue;
	while ((start < len) && ((val[start] & 0x000000ff) <= ' ')) {
		start++;
	}
	System.arraycopy(val, start, val, 0, intLength = len - start);
	return this;
}
/**
* Removes white space from the end of this ByteString. 
* <p>All characters that have codes less than or equal to <code>'&#92;u0020'</code> (the space character)
* are considered to be white space. 
* @return this ByteString
*/
public final ByteString trimRight() {
	int len = intLength;
	byte[] val = bytValue;
	while ((len > 0) && ((val[len - 1] & 0x000000ff) < 33)) {
		len--;
	}
	intLength = len;
	return this;
}
}