package net.conacom.servletrunner.util;

/**
 * Copyright 1999-2002 by conacom GmbH
 * Wilhelmshoeher Strasse 202, 60389 Frankfurt/Main, Germany
 * All rights reserved.
 * 
 * This software or implementation is free for commercial and non-commercial use.
 * No warranty is given about the functionality.
 */
import java.io.*;
import java.util.*;
import java.text.*;
/**
 * creates logfiles in the form {praefix}+yyyymmdd+{suffix}.
 * changes the logfile at midnight.
 * is able to dump stacktraces.
 * returns the formatted string for screen output.
 *
 * a message has the form "{ID} message" where ID is a single char. this char 
 * is matched against the filterstring (setFilter(...)).
 *
 * usually the following flags should be defined:
 * D = Debug
 * # = Informational messages
 * ! = Warnings
 * ? = Errors
 *
 * ** class is prepared to filter for screen and for file but the runner doesn't use it.
 *
 * output is "{ID} {Thread-Name} Message"
 *
 * 25.06.02 ho: minor fixes
 * 14.06.02 ho: warning about illegal logentry
 * 10.06.02 ho: if the filterstring is "", logging is switched off
 * 04.06.02 ho: getDateFormat(),getTimeFormat()
 * 24.05.02 ho: check NULL on pwLog before writing (.dispose() when crashed
 * 22.05.02 ho: writeMessage(Throwable), strCurrentLogfile
 * 08.05.02 ho: synchronize on System.out
 * 07.05.02 ho: filter classname from "S"ervlet entries.
 * 12.04.02 ho: performance update (openFile() with seconds past midnight), logfilter screen & file
 * 07.04.02 ho: System.currentTime -> TimeFactory.getTime()
 * 25.03.02 ho: fixed: path ="" causes error
 * 24.03.02 ho: removed use of CT tool
 * 20.03.02 ho: synchronisation in writeInternal
 */
public class LogContext {
	// path to the logfile (like c:\runner\logfiles\ or /usr/logfiles/)
	private String strFilePath = null;
	// praefix like "runner_"
	private String strFilePraefix = null;
	// suffix like ".log"
	private String strFileSuffix = null;
	// where to write the logfile
	private PrintWriter pwLog = null;
	// universal StringBuffer to avoid unnessesary object-creation
	private StringBuffer strb = null;
	// LineFeed
	private static String LF = null;
	// log-options to filter (for file)
	private String strLogFilter_File = null;
	// log-options to filter (for screen)
	private String strLogFilter_Screen = null;
	// universal DateFormat ** synchronize on this object if you're manipulating datBuffer or sdf !
	private SimpleDateFormat sdf = null;
	// pattern for date
	private String strDateFormat = "dd.MM.yyyy";
	// pattern for time
	private String strTimeFormat = "HH.mm.ss";
	// universal date
	private Date datBuffer = null;
	// last midnight (unix time)
	private long lngNextMidnight = -1;
	// remove classname from logentry (only for "S" entries)
	private boolean blnRemoveClassName = false;
	// the current logfilename
	private String strCurrentLogfile = null;
/**
 * opens a logwriter with a given path, praefix & suffix
 */
public LogContext(String strPath, String strPraefix, String strSuffix) throws IOException {
	super();
	this.strFilePath = strPath;
	this.strFilePraefix = strPraefix;
	this.strFileSuffix = strSuffix;
	init();
}
/**
 * 
 */
public void dispose() {
	if (pwLog != null) {
		pwLog.flush();
		pwLog.close();
	}
	pwLog = null;
	strLogFilter_File = null;
	strLogFilter_Screen = null;
	strFilePath = null;
	strFilePraefix = null;
	strFileSuffix = null;
	strb = null;
}
/**
 * returns the current logfilename
 */
public String getCurrentLogFile() {
	return strCurrentLogfile;
}
/**
 * gets the date format. see: SimpleDateFormat
 */
public String getDateFormat() {
	return strDateFormat;
}
/**
 * gets the time format. see: SimpleDateFormat
 */
public String getTimeFormat() {
	return strTimeFormat;
}
/**
 * init stuff
 */
private void init() throws IOException {
	//
	if (strFileSuffix == null)
		strFileSuffix = "";
	if (strFileSuffix.length() != 0 && !strFileSuffix.startsWith("."))
		strFileSuffix = "." + strFileSuffix;
	//
	if (strFilePath == null)
		strFilePath = "";
	if (strFilePath.length() != 0 && !strFilePath.endsWith(File.separator))
		strFilePath = strFilePath + File.separator;
	//
	LF = System.getProperty("line.separator");
	strb = new StringBuffer(100);
	sdf = new SimpleDateFormat();
	datBuffer = new Date();
	lngNextMidnight = -1;
	openFile(TimeFactory.getTime());
}
/**
 * opens a file if necessary. reopens it, if date has changed
 */
private void openFile(long lngSessionTime) throws IOException {
	// check if we crossed midnight
	if (lngSessionTime > lngNextMidnight) {
		lngNextMidnight = new java.sql.Date(TimeFactory.getTime()).getTime() + 86400000;
	} else {
		return;
	}
	// create date in format YYYYMMDD
	String strDate = null;
	synchronized (sdf) {
		datBuffer.setTime(lngSessionTime);
		sdf.applyPattern("yyyyMMdd");
		strDate = sdf.format(datBuffer);
	}
	// filename
	String strFile = strFilePraefix + strDate + strFileSuffix;
	// (re-)open the logfile
	if (pwLog != null) {
		pwLog.flush();
		pwLog.close();
	}
	// create the directory if needed
	if (strFilePath.length() != 0) {
		File path = new File(strFilePath);
		if (!path.exists()) {
			path.mkdirs();
		}
	}
	// build the filename (with path)
	if (strFilePath.length() != 0) {
		if (strFilePath.endsWith(File.separator)) {
			strFile = strFilePath + strFile;
		} else {
			strFile = strFilePath + File.separator + strFile;
		}
	}
	// open the logfile (append, autoflush)
	pwLog = new PrintWriter(new FileWriter(strFile, true), true);
	strCurrentLogfile = strFile;
}
/**
 * sets the date format. see: SimpleDateFormat
 */
public void setDateFormat(String strPattern) {
	strDateFormat = strPattern;
}
/**
 * sets the filter for file logging
 */
public void setFilter_File(String strFilter) {
	strLogFilter_File = strFilter;
}
/**
 * sets the filter for screen logging
 */
public void setFilter_Screen(String strFilter) {
	strLogFilter_Screen = strFilter;
}
/**
 * shall we remove classnames from "S"ervlet logentries ?
 */
public void setRemoveClassName(boolean b) {
	blnRemoveClassName = b;
}
/**
 * sets the time format. see: SimpleDateFormat
 */
public void setTimeFormat(String strPattern) {
	strTimeFormat = strPattern;
}
/**
 * checks if a string should be logged
 */
private boolean validFilter(String strLogLine, String strMask) {
	if (strMask == null)
		return true;
	if (strMask.length() == 0)
		return false;
	return (strMask.indexOf(strLogLine) != -1);
}
/**
 * checks if a string should be logged to file
 */
private boolean validFilter_File(String s) {
	return validFilter(s, strLogFilter_File);
}
/**
 * checks if a string should be logged to screen
 */
private boolean validFilter_Screen(String s) {
	return validFilter(s, strLogFilter_Screen);
}
public void write(String strMessage) {
	writeInternal(strMessage, null, TimeFactory.getTime());
}
public void write(String strMessage, Throwable t) {
	writeInternal(strMessage, t, TimeFactory.getTime());
}
public void write(Throwable t) {
	writeInternal(null, t, TimeFactory.getTime());
}
/**
 * writes to the log. returns the formatted logfile-entry also (for logging to screen)
 *
 * strMsg    = the message to log. Format "* Message" where "*" is char of "D!?#"
 * thr       = an exception to log
 * lngTime   = the timestamp for this entry or stacktrace
 *
 */
private void writeInternal(String strMsg, Throwable thr, long lngTime) {
	if (strMsg == null) {
		strMsg = "! " + thr.getMessage();
	}
	/**
	 * check if this entry 
	 */
	String strPraefix = strMsg.substring(0, 1);
	if (strPraefix.equals(" ")) {
		write("! Warning: invalid logentry follows");
	}
	String strSuffix = strMsg.substring(1).trim();
	if (!(validFilter_Screen(strPraefix) || validFilter_File(strPraefix)))
		return;

	/**
	 * filter classname from logstring
	 */
	if (blnRemoveClassName && strPraefix.equals("S")) {
		int pp = strSuffix.indexOf(":");
		if (pp != -1)
			strSuffix = strSuffix.substring(pp + 1).trim();
	}

	/**
	 * create the logstring
	 */
	strb.setLength(0);
	synchronized (sdf) {
		datBuffer.setTime(lngTime);
		sdf.applyPattern(strDateFormat);
		strb.append(sdf.format(datBuffer));
		strb.append(" ");
		sdf.applyPattern(strTimeFormat);
		strb.append(sdf.format(datBuffer));
	}
	String strMS = Long.toString(lngTime);
	strb.append(".");
	strb.append(strMS.substring(strMS.length() - 4));
	strb.append(" ");
	strb.append(strPraefix);
	strb.append(" '");
	strb.append(Thread.currentThread().getName());
	strb.append("' ");
	String strLog = strb.toString();
	//
	strb.setLength(0);
	strb.append(strLog);
	strb.append(strSuffix);
	//
	if (thr != null) {
		try {
			StringWriter sw = new StringWriter();
			PrintWriter pw = new PrintWriter(sw);
			thr.printStackTrace(pw);
			BufferedReader br = new BufferedReader(new StringReader(sw.toString()));
			String strBuffer = null;
			while ((strBuffer = br.readLine()) != null) {
				strBuffer = strBuffer.trim();
				if (strBuffer.length() != 0) {
					strb.append(LF);
					strb.append(strLog);
					strb.append("::");
					strb.append(strBuffer.trim());
				}
			}
		} catch (IOException e) {
			// never happens because we're I/O'ing strings only
		}
	}
	// now write the entry
	if (validFilter_File(strPraefix) && pwLog != null) {
		synchronized (pwLog) {
			try {
				openFile(lngTime);
				pwLog.println(strb.toString());
			} catch (Exception e) {
				System.out.println("LogWriter: " + e.getMessage());
			}
		}
	}
	if (validFilter_Screen(strPraefix)) {
		synchronized (System.out) {
			System.out.println(strb.toString());
		}
	}
}
}