package net.conacom.servletrunner.util;

/**
 * Copyright 1999-2002 by conacom GmbH
 * Wilhelmshoeher Strasse 202, 60389 Frankfurt/Main, Germany
 * All rights reserved.
 * 
 * This software or implementation is free for commercial and non-commercial use.
 * No warranty is given about the functionality.
 */
import java.lang.reflect.*;
import javax.servlet.ServletException;
/**
 * allows monitored execution of a class
 *
 * 1) Object objResult = monitor.execute(.....)
 * 2) if (monitor.hadTimeout()) -> timeed out
 * 3) if (monitor.getThrowable()) -> had error
 * 4) otherwise everything was OK and the result (if any) is in objResult
 *
 * 16.06.02 ho: we are caching the method now. 3 times faster execution.
 * 15.06.02 ho: we are reusing the thread now. 4 times faster execution.
 * 22.05.02 ho: returns rootcause for ServletExceptions
 * 21.05.02 ho: return target exception in case of InvocationTargetException; threadname
 * 19.05.02 ho: fix: assume timeout in .run()
 * 15.05.02 ho: bugfixing, bugfixing, bugfixing
 * 14.05.02 ho: creation
 */
public class MonitoredExecution implements Runnable {
	// the class where we shall call the method
	private Object objClass = null;
	// the name of the method
	private String strMethod = null;
	// the argument values
	private Object objArgs[] = null;
	// the argument types
	private Class clsArgs[] = null;
	// the method to call
	private Method mthMother = null;
	// the result
	private Object objResult = null;
	// the execution thread
	private Thread thrd = null;
	// an error occured during execution of the method ?
	private Throwable thrwError = null;
	// execution timeout ?
	private volatile boolean blnHadTimeout = false;
	// lock/notify object main -> thread
	private LockObject objNotifyThread = null;
	// lock/notify object thread -> main
	private LockObject objNotifyMain = null;
	// instance do be disposed ?
	private volatile boolean blnStopThread = false;
	// cache..
	private boolean blnUseCache = false;
	private Object objCacheClass = null;
	private String strCacheMethod = null;
	private Class clsCacheArgs[] = null;
	// name of the parent thread
	private String strParentThreadName = null;
	// flag: currently executing something
	private boolean blnExecuting = false;
/**
 * default constructor with caching enabled
 */
public MonitoredExecution() {
	this(true);
}
/**
 * you can switch of caching here. use FALSE as parameter if your method 
 * to monitor changes between every call.
 */
public MonitoredExecution(boolean blnUseCache) {
	super();
	this.blnUseCache = blnUseCache;
	init();
}
/**
 * compares the cached caller structure with the new caller structure. true: both are equal.
 */
private boolean cacheValid() {
	// no caching ?
	if (!blnUseCache)
		return false;
	// 0) first call ?
	if (objCacheClass == null)
		return false;
	// 1) check for different caller
	if (objCacheClass != objClass)
		return false;
	// 2) check number of parameters 
	if (clsCacheArgs.length != clsArgs.length)
		return false;
	// 3) type of parameters
	for (int i = 0; i < clsCacheArgs.length; i++) {
		if (!clsCacheArgs[i].equals(clsArgs[i]))
			return false;
	}
	// 4) methodname
	if (!strCacheMethod.equals(strMethod))
		return false;
	//
	return true;
}
/**
 * cleanup, release ressources
 */
public void dispose() {
	if (thrd != null) {
		blnStopThread = true;
		objNotifyThread.lo_notify();
		try {
			thrd.join(1000);
		} catch (InterruptedException e) {
		}
		if (thrd.isAlive()) {
			thrd.stop();
		}
		thrd = null;
	}
	objNotifyMain.dispose();
	objNotifyMain = null;
	objNotifyThread.dispose();
	objNotifyThread = null;
	objClass = null;
	objArgs = null;
	clsArgs = null;
	mthMother = null;
	objResult = null;
	thrwError = null;
}
/**
 * executes the method of a class.
 *
 * returns an Object with the resultvalue of the method (if any) or NULL if there wasn't any returncode
 */
public Object execute(Object objClass, String strMethod, Object objArgs[], Class clsArgs[], long lngTimeout) {
	blnExecuting = true;
	this.objClass = objClass;
	this.objArgs = objArgs;
	this.clsArgs = clsArgs;
	this.strMethod = strMethod;

	/**
	 * sanity check
	 */
	if (thrd == null || thrd.isAlive() == false) {
		throw new net.conacom.servletrunner.main.RunnerException("? Thread isn't ready !");
	}

	/**
	 * init
	 */
	thrwError = null;
	blnHadTimeout = false;

	/**
	 * test if we can use the method from the previous call. otherwise find the method to call.
	 */
	if (!cacheValid()) {
		try {
			mthMother = objClass.getClass().getMethod(strMethod, clsArgs);
		} catch (NoSuchMethodException e) {
			thrwError = new net.conacom.servletrunner.main.RunnerException("No such method '" + strMethod + "'");
			return null;
		}
		setCache();
	}

	/**
	 * set name (for logging) and notify the thread to start.
	 */
	objNotifyMain.lo_reset(); // clear state to be safe
	objNotifyThread.lo_notify(); // notify thread

	/**
	 * wait for a signal of the thread.
	 */
	objNotifyMain.lo_wait(lngTimeout);

	/**
	 * if the execute call didn't finish in time, stop the thread (and create a new thread).
	 */
	if (blnHadTimeout) {
		thrd.stop();
		thrd = null;
		objNotifyThread.lo_reset();
		threadStart();
	}

	/**
	 * cleanup & exit
	 */
	blnExecuting = false;
	return objResult;
}
/**
 * gets the error (or NULL)
 */
public Throwable getThrowable() {
	return thrwError;
}
/**
 * returns TRUE if the execuction of the method timed out.
 */
public boolean hadTimeout() {
	return blnHadTimeout;
}
/**
 * initializes the tool and starts the first thread
 */
private void init() {
	objNotifyMain = new LockObject();
	objNotifyThread = new LockObject();
	strParentThreadName = Thread.currentThread().getName();
	threadStart();
}
/**
 * returns true if the monitor is executing something
 */
public boolean isExecuting() {
	return blnExecuting;
}
/**
 * executes a method
 */
public void run() {
	while (true) {
		/**
		 * init
		 */
		thrwError = null;

		/**
		 * wait for wakeup. exit, if the thread should be shutdown normally
		 */
		objNotifyThread.lo_wait(0);
		if (blnStopThread) {
			return;
		}

		/**
		 * assume that we will timeout
		 */
		blnHadTimeout = true;

		/**
		 * execute the method. catch thread death separatly (rises when main method calls .stop() on the object)
		 */
		String strOldThreadName = thrd.getName();
		try {
			thrd.setName(strParentThreadName);
			objResult = mthMother.invoke(objClass, objArgs);
		} catch (InvocationTargetException e) {
			thrwError = e.getTargetException();
			// get rootcause of ServletExceptions
			if (thrwError instanceof ServletException) {
				ServletException ex = (ServletException) thrwError;
				if (ex.getRootCause() != null)
					thrwError = ex.getRootCause();
			}
		} catch (Throwable e) {
			thrwError = e;
		}
		thrd.setName(strOldThreadName);

		/**
		 * .stop() called ? end thread by rethrowing the error.
		 */
		if (thrwError instanceof ThreadDeath) {
			thrwError = null;
			throw new ThreadDeath();
		}

		/**
		 * normal end. mark it and notify the waiting main method
		 */
		blnHadTimeout = false;
		objNotifyMain.lo_notify();
	}
}
/**
 * 
 */
private void setCache() {
	// no caching ?
	if (!blnUseCache)
		return;
	//
	objCacheClass = objClass;
	strCacheMethod = strMethod;
	clsCacheArgs = new Class[clsArgs.length];
	for (int i = 0; i < clsCacheArgs.length; i++) {
		clsCacheArgs[i] = clsArgs[i];
	}
}
/**
 * 
 */
private void threadStart() {
	blnStopThread = false;
	if (thrd != null)
		return;
	thrd = new Thread(this, "MonitoredExecution for " + Thread.currentThread().getName());
	thrd.start();
}
}