package net.conacom.servletrunner.util;

/**
 * Copyright 1999-2002 by conacom GmbH
 * Wilhelmshoeher Strasse 202, 60389 Frankfurt/Main, Germany
 * All rights reserved.
 * 
 * This software or implementation is free for commercial and non-commercial use.
 * No warranty is given about the functionality.
 */
import java.io.*;
import java.net.*;
import java.util.*;
import java.util.zip.*;
import net.conacom.servletrunner.main.*;
/**
 * a classloader which can load classes from another directory
 * the systemclasspath is searched first
 *
 * we don't keep the archives open. this slows down loading, but in this way we can exchange the 
 * archive during runtime and restart the runner.
 *
 * 11.06.02 ho: added local path to container classpath
 * 23.05.02 ho: getResource(..) and getResourceAsStream(..) are removing leading / from 
 *				the file now. posibility to add an additional, temporary classpath (ServletContext)
 * 11.05.02 ho: bug: loadFromStream(..) gots file-formatted string instead of the original classname
 * 07.05.02 ho: avoid NullPointerException if the classpath contains an invalid entry
 * 28.04.02 ho: getResourceAsStream(..), ServletContext directory
 * 26.03.02 ho: creation
 *
 * to do: merge getResourceAsStream and loadFrom*
 *
 */
public class RunnerClassLoader extends ClassLoader {
	// separator for the userdefined classpath
	private final static String CLASSPATH_SEPARATOR = ";";
	// vector of all classpath elements (path1;path2;path3;)
	private Vector vecClassPath = null;
	// root of the servletcontext
	private String strContextRoot = "";
	// classcache
	private Hashtable hshClassCache = null;
	// where to log
	private LogContext log = null;
	// local class for storing informations about the classpath elements
	private class ClassPathEntry {
		// the path
		private String strPath;
		// does the classpath points to an archive (.jar, .zip) ?
		private boolean blnIsArchive = false;
		// reference to a zip-file for closing it in .dispose();
		private ZipFile zf = null;
		// string representation of this object
		public String toString() {
			return "[Path:" + strPath + ",Archive:" + String.valueOf(blnIsArchive) + "]";
		}
	}
public RunnerClassLoader(String strClassPath, String strContextRoot, LogContext log) {
	super();
	this.log = log;
	this.strContextRoot = strContextRoot;
	init0(strClassPath);
}
/**
 * clones the classpath and inserts a classpath at the begin
 */
private Vector appendClassPath(String str) {
	Vector v = (Vector) vecClassPath.clone();
	if (str.length() > 0) {
		str = Helper.platformPath(str);
		if (!str.endsWith(File.separator))
			str = str + File.separator;
		ClassPathEntry cp = new ClassPathEntry();
		cp.strPath = str;
		cp.blnIsArchive = false;
		v.insertElementAt(cp, 0);
	}
	return v;
}
/**
 * 
 */
public void dispose() {
	log.write("C Disposing Classloader");
	if (vecClassPath != null) {
		for (int i = 0; i < vecClassPath.size(); i++) {
			ClassPathEntry cp = (ClassPathEntry) vecClassPath.elementAt(i);
			// we have to close the zipfile because maybe the user forgot
			// to close the inputstream from "getResourceAsStream()"
			if (cp.zf != null) {
				Helper.close(cp.zf);
			}
		}
		vecClassPath.removeAllElements();
		vecClassPath = null;
	}
	if (hshClassCache != null) {
		hshClassCache.clear();
		hshClassCache = null;
	}
}
/**
 * tries to get the resource from system or the local classpath
 * WARNING: JDK 1 doesn't support "jar://" urls !
 */
public URL getResource(String strFile) {
	return getResource(strFile, "");
}
/**
 * tries to get the resource from system or the local classpath
 * WARNING: JDK 1 doesn't support jar: urls !
 */
public URL getResource(String strFile, String strAddClassPath) {
	strFile = Helper.platformPath(strFile);
	if (strFile.startsWith(File.separator))
		strFile = strFile.substring(1);
	URL url = null;
	// load from system
	url = super.getSystemResource(strFile);
	if (url != null)
		return url;
	// create a new, temporary classpath
	Vector v = appendClassPath(strAddClassPath);
	// load from directory or zip
	for (int i = 0; i < v.size(); i++) {
		ClassPathEntry cp = (ClassPathEntry) v.elementAt(i);
		if (cp.blnIsArchive) {
			ZipFile zf = null;
			try {
				zf = new ZipFile(new File(cp.strPath));
				ZipEntry ze = zf.getEntry(strFile);
				if (ze != null) {
					url = new URL("jar:" + cp.strPath + "!/" + strFile);
				}
			} catch (ZipException e) {
			} catch (MalformedURLException e) {
				log.write("Bad URL", e);
			} catch (IOException e) {
			} finally {
				Helper.close(zf);
			}
		} else {
			try {
				url = new URL("file:" + cp.strPath + strFile);
			} catch (MalformedURLException e) {
				log.write("Bad URL", e);
			}
		}
	}
	//
	return url;
}
/**
 * tries to open a resource from system or the local classpath
 */
public InputStream getResourceAsStream(String strFile) {
	return getResourceAsStream(strFile, "");
}
/**
 * tries to open a resource from system or the local classpath
 */
public InputStream getResourceAsStream(String strFile, String strAddClassPath) {
	strFile = Helper.platformPath(strFile);
	if (strFile.startsWith(File.separator))
		strFile = strFile.substring(1);
	InputStream in = null;
	// load from system
	in = super.getSystemResourceAsStream(strFile);
	if (in != null)
		return in;
	// create a new, temporary classpath
	Vector v = appendClassPath(strAddClassPath);
	// load from directory or zip
	for (int i = 0; i < v.size(); i++) {
		ClassPathEntry cp = (ClassPathEntry) v.elementAt(i);
		if (cp.blnIsArchive) {
			ZipFile zf = null;
			try {
				zf = new ZipFile(new File(cp.strPath));
				cp.zf = zf;
				ZipEntry ze = zf.getEntry(strFile);
				if (ze != null) {
					in = zf.getInputStream(ze);
				}
			} catch (ZipException e) {
				Helper.close(in);
				Helper.close(zf);
			} catch (IOException e) {
				Helper.close(in);
				Helper.close(zf);
			}
		} else {
			try {
				in = new FileInputStream(new File(cp.strPath, strFile));
			} catch (IOException e) {
				Helper.close(in);
			}
		}
		if (in != null)
			break;
	}
	//
	return in;
}
/**
 * performs local initializations
 */
private void init0(String strClassPath) {

	/**
	 * parse the classpath into the vector of classes
	 */
	// add the contextroot to the classpath
	if (strContextRoot.length() != 0)
		strClassPath = strContextRoot + CLASSPATH_SEPARATOR + strClassPath;
	//
	// add the classpath of the VM because there may be other classes in the commandline
	//strClassPath = strClassPath + CLASSPATH_SEPARATOR + System.getProperty("java.class.path");

	// add ; to the end for better parsing
	if (!strClassPath.endsWith(CLASSPATH_SEPARATOR))
		strClassPath = strClassPath + CLASSPATH_SEPARATOR;
	//
	vecClassPath = new Vector(1, 1);
	//
	StringTokenizer strt = new StringTokenizer(strClassPath, CLASSPATH_SEPARATOR);
	while (strt.hasMoreTokens()) {
		ClassPathEntry cpe = new ClassPathEntry();
		String strFile = strt.nextToken();
		String strDir = Helper.getDirectory(strFile);
		// check if the classpathentry is a directory or a file
		if (strDir == null) {
			log.write("Classpath " + strFile + " isn't a valid directory or file, ignored");
		} else {
			// the context-root is the directory where the configfile of this container resides.
			// add context-root to string if string isn't a directory.
			if (strDir.length() == 0)
				strFile = strContextRoot + strFile;
			cpe.strPath = strFile;
			cpe.blnIsArchive = false;
			vecClassPath.addElement(cpe);
		}
	}

	/**
	 * no explicit classpath ? add at least the local directory
	 */
	if (vecClassPath.size() == 0) {
		ClassPathEntry cpe = new ClassPathEntry();
		cpe.strPath = ".";
		cpe.blnIsArchive = false;
		vecClassPath.addElement(cpe);
	}

	/**
	 * first check, if the new classpath is OK
	 */
	for (int i = 0; i < vecClassPath.size(); i++) {
		ClassPathEntry cpe = (ClassPathEntry) vecClassPath.elementAt(i);
		File f = new File(cpe.strPath);
		if (!f.exists()) {
			throw new RunnerException(f.toString() + " is not a valid file/directory");
		}
		if (!f.canRead()) {
			throw new RunnerException(f.toString() + " is not readable");
		}
		// check if its a valid zipfile and mark the result
		if (f.isFile()) {
			try {
				ZipFile zf = new ZipFile(f);
				zf.close();
				cpe.blnIsArchive = true;
			} catch (Throwable t) {
				cpe.blnIsArchive = false;
			}
		}
		log.write("C Added " + cpe.toString() + " to container classpath");
	}

	/**
	 * create the class cache
	 */
	hshClassCache = new Hashtable();
}
/**
 * Resolves the specified name to a Class. The method loadClass() is 
 * called by the virtual machine.
 * <p>
 * If the <code>resolve</code> flag is true, the method should call 
 * the <code>resolveClass</code> method on the resulting class object.
 * <p>
 * As an abstract method, loadClass() must be defined in a subclass of 
 * ClassLoader. By using a Hashtable, you can avoid loading the same 
 * Class more than once.
 * 
 * @param	   name     the name of the desired Class.
 * @param      resolve  true if the Class needs to be resolved.
 * @return	   the resulting Class, or null if it was not found.
 * @exception  ClassNotFoundException  if the class loader cannot find
 *               a definition for the class.
 * @see	   java.util.Hashtable
 * @since      JDK1.0
 */
protected Class loadClass(String strClass, boolean blnResolve) throws ClassNotFoundException {

	/**
	 * class already in the cache ?
	 */
	Class cls = (Class) hshClassCache.get(strClass);
	if (cls != null)
		log.write("C Loaded " + strClass + " from cache");

	/**
	 * try to load the class from the system
	 */
	if (cls == null) {
		try {
			cls = findSystemClass(strClass);
		} catch (Throwable t) {
			cls = null;
		}
		if (cls != null)
			log.write("C Loaded " + strClass + " from systempath");
	}

	/** 
	 * try to load the class from our local classpath
	 */
	if (cls == null) {
		for (int i = 0; i < vecClassPath.size(); i++) {
			ClassPathEntry cp = (ClassPathEntry) vecClassPath.elementAt(i);
			if (cp.blnIsArchive) {
				cls = loadFromZip(strClass, cp.strPath);
			} else {
				cls = loadFromFile(strClass, cp.strPath);
			}
			if (cls != null) {
				log.write("C Loaded " + strClass + " from containerpath");
				break;
			}
		}
	}

	/**
	 * class not found or class not loadable ?
	 */
	if (cls == null) {
		hshClassCache.remove(strClass); // to be sure
		log.write("? Cannot find class " + strClass);
		throw new ClassNotFoundException(strClass);
	}

	/**
	 * we are done
	 */
	if (blnResolve)
		resolveClass(cls);
	//
	hshClassCache.put(strClass, cls);
	//
	return cls;
}
/**
 * tries to load a class from a file
 * returns NULL if the class cannot be found or something else went wrong
 */
private Class loadFromFile(String strClass, String strPath) {
	//
	String strFile = strClass.replace('.', File.separatorChar);
	strFile = strFile + ".class";
	//
	while (!Character.isJavaIdentifierStart(strFile.charAt(0)))
		strFile = strFile.substring(1);
	//
	InputStream in = null;
	Class cls = null;
	try {
		in = new FileInputStream(new File(strPath, strFile));
		cls = loadFromStream(strClass, in);
	} catch (IOException e) {
	} finally {
		Helper.close(in);
	}
	return cls;
}
/**
 * tries to load a class from an inputstream
 * returns NULL if something went wrong
 */
private Class loadFromStream(String strClass, InputStream in) {
	ByteString bs = new ByteString(8192);
	byte buffer[] = new byte[512];
	int i = -1;
	try {
		while ((i = in.read(buffer)) != -1) {
			bs.append(buffer, 0, i);
		}
	} catch (IOException e) {
		return null;
	}
	//
	if (bs.getLength() == 0)
		return null;
	return defineClass(strClass, bs.getByteArray(), 0, bs.getLength());
}
/**
 * tries to load a class from a zipfile.
 * returns NULL if the class cannot be found or something else went wrong
 */
private Class loadFromZip(String strClass, String strArchive) {
	//
	String strFile = strClass.replace('.', '/');
	strFile = strFile + ".class";
	//
	ZipFile zf = null;
	InputStream in = null;
	Class cls = null;
	try {
		zf = new ZipFile(new File(strArchive));
		ZipEntry ze = zf.getEntry(strFile);
		if (ze != null) {
			in = zf.getInputStream(ze);
			cls = loadFromStream(strClass, in);
		}
	} catch (ZipException e) {
	} catch (IOException e) {
	} finally {
		Helper.close(in);
		Helper.close(zf);
	}
	return cls;
}
}